# A convenience module for getting boa stuff easily
from boa.contracts.vyper.vyper_contract import (  # noqa: F401
    VyperContract,
    VyperDeployer,
)
from boa_zksync.contract import ZksyncContract  # noqa: F401
from boa_zksync.deployer import ZksyncDeployer  # noqa: F401



from pathlib import Path

from boa.deployments import DeploymentsDB

from moccasin.config import Config


def test_generate_sql_from_args_with_where(blank_tempdir):
    # Arrange
    db_path = blank_tempdir + "/.deployments.db"
    db = DeploymentsDB(db_path)
    contract_name = "MockV3Aggregator"
    chain_id = 31337
    limit = 1

    config = Config(Path("."))
    # This should be a blank pyevm network
    active_network = config.get_active_network()

    # Act
    sql_query, params = active_network._generate_sql_from_args(
        contract_name=contract_name, chain_id=chain_id, limit=limit, db=db
    )
    expected_sql = "SELECT contract_address,contract_name,rpc,deployer,tx_hash,broadcast_ts,tx_dict,receipt_dict,source_code,abi,session_id,deployment_id FROM deployments WHERE contract_name = ? AND json_extract(tx_dict, '$.chainId') = ? ORDER BY broadcast_ts DESC LIMIT ? "
    expected_parametrs = ("MockV3Aggregator", "31337", 1)

    # Assert
    assert sql_query == expected_sql
    assert params == expected_parametrs


def test_generate_sql_from_args_without_where(blank_tempdir):
    # Arrange
    db_path = blank_tempdir + "/.deployments.db"
    db = DeploymentsDB(db_path)
    config = Config(Path("."))
    # This should be a blank pyevm network
    active_network = config.get_active_network()

    # Act
    sql_query, params = active_network._generate_sql_from_args(db=db)
    expected_sql = "SELECT contract_address,contract_name,rpc,deployer,tx_hash,broadcast_ts,tx_dict,receipt_dict,source_code,abi,session_id,deployment_id FROM deployments ORDER BY broadcast_ts DESC "
    expected_parametrs = ()

    # Assert
    assert sql_query == expected_sql
    assert params == expected_parametrs


def test_generate_sql_from_args_without_and(blank_tempdir):
    # Arrange
    db_path = blank_tempdir + "/.deployments.db"
    db = DeploymentsDB(db_path)
    contract_name = "MockV3Aggregator"
    limit = 1
    config = Config(Path("."))
    # This should be a blank pyevm network
    active_network = config.get_active_network()

    # Act
    sql_query, params = active_network._generate_sql_from_args(
        contract_name=contract_name, limit=limit, db=db
    )
    expected_sql = "SELECT contract_address,contract_name,rpc,deployer,tx_hash,broadcast_ts,tx_dict,receipt_dict,source_code,abi,session_id,deployment_id FROM deployments WHERE contract_name = ? ORDER BY broadcast_ts DESC LIMIT ? "
    expected_parametrs = ("MockV3Aggregator", 1)

    # Assert
    assert sql_query == expected_sql
    assert params == expected_parametrs



import os
import subprocess
from pathlib import Path

from moccasin.config import Config
from moccasin.constants.vars import (
    DEPENDENCIES_FOLDER,
    GITHUB,
    PACKAGE_VERSION_FILE,
    PYPI,
)
from tests.conftest import (
    patrick_org_name,
    patrick_package_name,
    pip_package_name,
    version,
)


def test_can_purge_github_no_version(
    purge_temp_path, purge_reset_dependencies, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(purge_temp_path)
        result = subprocess.run(
            [mox_path, "purge", f"{patrick_package_name}"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert f"Removed {patrick_package_name}" in result.stderr
    project_root: Path = Config.find_project_root(Path(purge_temp_path))
    config = Config(project_root)
    assert f"{patrick_package_name}@{version}" not in config.dependencies
    assert (
        not Path(purge_temp_path)
        .joinpath(f"{DEPENDENCIES_FOLDER}/{GITHUB}/{patrick_org_name}")
        .exists()
    )
    versions_file = (
        purge_temp_path.joinpath(DEPENDENCIES_FOLDER)
        .joinpath(GITHUB)
        .joinpath(PACKAGE_VERSION_FILE)
    )
    with open(versions_file) as f:
        assert f.read() == ""


def test_can_purge_github_with_version(
    purge_temp_path, purge_reset_dependencies, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(purge_temp_path)
        result = subprocess.run(
            [mox_path, "purge", f"{patrick_package_name}@{version}"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert f"Removed {patrick_package_name}" in result.stderr
    project_root: Path = Config.find_project_root(Path(purge_temp_path))
    config = Config(project_root)
    assert f"{patrick_package_name}@{version}" not in config.dependencies
    assert (
        not Path(purge_temp_path)
        .joinpath(f"{DEPENDENCIES_FOLDER}/{GITHUB}/{patrick_org_name}")
        .exists()
    )
    versions_file = (
        purge_temp_path.joinpath(DEPENDENCIES_FOLDER)
        .joinpath(GITHUB)
        .joinpath(PACKAGE_VERSION_FILE)
    )
    with open(versions_file) as f:
        assert f.read() == ""


def test_can_purge_pip(purge_temp_path, purge_reset_dependencies, mox_path):
    current_dir = Path.cwd()
    try:
        os.chdir(purge_temp_path)
        result = subprocess.run(
            [mox_path, "purge", pip_package_name],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert f"Removed {pip_package_name}" in result.stderr
    project_root: Path = Config.find_project_root(Path(purge_temp_path))
    config = Config(project_root)
    assert pip_package_name not in config.dependencies
    assert (
        not Path(purge_temp_path)
        .joinpath(f"{DEPENDENCIES_FOLDER}/{PYPI}/{pip_package_name}")
        .exists()
    )



import os
from pathlib import Path

from moccasin.commands.inspect import inspect_contract
from tests.conftest import COMPLEX_PROJECT_PATH


def test_inspect_counter(complex_project_config):
    expected_dir = {
        "set_number(uint256)": "0xd6d1ee14 (3604082196)",
        "increment()": "0xd09de08a (3500007562)",
    }
    current_dir = Path.cwd()
    try:
        os.chdir(current_dir.joinpath(COMPLEX_PROJECT_PATH))
        result = inspect_contract("Counter", "function_signatures", print_out=False)
    finally:
        os.chdir(current_dir)
    assert result == expected_dir



import os
from pathlib import Path

from moccasin.commands.run import get_script_path, run_script
from tests.conftest import COMPLEX_PROJECT_PATH


def test_get_script_path_for_single_name(complex_project_config):
    script_path = get_script_path("deploy")
    assert script_path == complex_project_config.get_root() / "script/deploy.py"


def test_get_script_path_for_relative_path(complex_project_config):
    script_path = get_script_path("./script/deploy.py")
    assert script_path == complex_project_config.get_root() / "script/deploy.py"


def test_get_script_path_for_absolute_path(complex_temp_path, complex_project_config):
    script_path = get_script_path(complex_temp_path.joinpath("./script/deploy.py"))
    assert script_path == complex_project_config.get_root() / "script/deploy.py"


def test_no_prompt_on_test_networks(complex_project_config, capsys):
    current_dir = Path.cwd()
    try:
        os.chdir(COMPLEX_PROJECT_PATH)
        run_script("deploy", network="pyevm")
    finally:
        os.chdir(current_dir)
    captured = capsys.readouterr()
    assert "Starting count:  0" in captured.out
    assert "Ending count:  1" in captured.out



import os
import re
import shutil
import subprocess
import sys
import tomllib
import traceback
import zipfile
from argparse import Namespace
from base64 import b64encode
from io import BytesIO
from pathlib import Path
from urllib.parse import quote

import requests  # type: ignore
import tomli_w
from tqdm import tqdm

from moccasin._dependency_utils import (
    DependencyType,
    _write_new_dependencies,
    classify_dependency,
)
from moccasin.config import get_or_initialize_config
from moccasin.constants.vars import GITHUB, PACKAGE_VERSION_FILE, PYPI, REQUEST_HEADERS
from moccasin.logging import logger


def main(args: Namespace):
    requirements = args.requirements
    config = get_or_initialize_config()
    if len(requirements) == 0:
        requirements = config.get_dependencies()
    if len(requirements) == 0:
        logger.info("No dependencies to install.")
        return 0

    pip_requirements = []
    github_requirements = []
    for requirement in requirements:
        if classify_dependency(requirement) == DependencyType.GITHUB:
            github_requirements.append(requirement)
        else:
            pip_requirements.append(requirement)
    install_path: Path = config.get_base_dependencies_install_path()
    if len(pip_requirements) > 0:
        _pip_installs(pip_requirements, install_path.joinpath(PYPI), args.quiet)
    if len(github_requirements) > 0:
        _github_installs(github_requirements, install_path.joinpath(GITHUB), args.quiet)
    return 0


# Much of this code thanks to brownie
# https://github.com/eth-brownie/brownie/blob/master/brownie/_config.py
def _github_installs(
    github_ids: list[str], base_install_path: Path, quiet: bool = False
):
    logger.info(f"Installing {len(github_ids)} GitHub packages...")
    for package_id in github_ids:
        try:
            if "@" in package_id:
                path, version = package_id.split("@", 1)
            else:
                path = package_id
                version = None  # We'll fetch the latest version later
            org, repo = path.split("/")
            org = org.strip().lower()
            repo = repo.strip().lower()
        except ValueError:
            raise ValueError(
                "Invalid package ID. Must be given as ORG/REPO[@VERSION]"
                "\ne.g. 'pcaversaccio/snekmate@v2.5.0'"
            ) from None

        headers = REQUEST_HEADERS.copy()
        headers.update(_maybe_retrieve_github_auth())

        if version is None:
            version = _get_latest_version(org, repo, headers)
            logger.info(f"Using latest version for {org}/{repo}: {version}")

        org_install_path = base_install_path.joinpath(f"{org}")
        org_install_path.mkdir(exist_ok=True, parents=True)
        repo_install_path = org_install_path.joinpath(f"{repo}")
        versions_install_path = base_install_path.joinpath(PACKAGE_VERSION_FILE)

        if repo_install_path.exists():
            with open(versions_install_path, "rb") as f:
                versions = tomllib.load(f)
                versions = {k.lower(): v for k, v in versions.items()}
                installed_version = versions.get(f"{org}/{repo}", None)
            if installed_version == version:
                logger.info(f"{org}/{repo} already installed at version {version}")
                return f"{org}/{repo}@{version}"
            else:
                logger.info(
                    f"Updating {org}/{repo} from {installed_version} to {version}"
                )

        if re.match(r"^[0-9a-f]+$", version):
            download_url = (
                f"https://api.github.com/repos/{org}/{repo}/zipball/{version}"
            )
        else:
            download_url = _get_download_url_from_tag(org, repo, version, headers)
        existing = list(repo_install_path.parent.iterdir())

        # Some versions contain special characters and github api seems to display url without
        # encoding them.
        # It results in a ConnectionError exception because the actual download url is encoded.
        # In this case we try to sanitize the version in url and download again.
        try:
            _stream_download(download_url, str(repo_install_path.parent), headers)
        except ConnectionError:
            download_url = f"https://api.github.com/repos/{org}/{repo}/zipball/refs/tags/{quote(version)}"
            _stream_download(download_url, str(repo_install_path.parent), headers)

        try:
            installed = next(
                i for i in repo_install_path.parent.iterdir() if i not in existing
            )
        except StopIteration:
            installed = None

        if installed:
            if repo_install_path.exists():
                shutil.rmtree(repo_install_path)
            shutil.move(installed, repo_install_path)

        # Update versions file
        if versions_install_path.exists():
            with open(versions_install_path, "rb") as f:
                versions_data = tomllib.load(f)
            versions_data[f"{org}/{repo}"] = version
            with open(versions_install_path, "wb") as f:
                tomli_w.dump(versions_data, f)
        else:
            with open(versions_install_path, "w", encoding="utf-8") as f:
                toml_string = tomli_w.dumps({f"{org}/{repo}": version})
                f.write(toml_string)
        _write_new_dependencies(github_ids, DependencyType.GITHUB)


def _get_latest_version(org: str, repo: str, headers: dict) -> str:
    response = requests.get(
        f"https://api.github.com/repos/{org}/{repo}/releases/latest", headers=headers
    )
    if response.status_code == 200:
        return response.json()["tag_name"].lstrip("v")

    response = requests.get(
        f"https://api.github.com/repos/{org}/{repo}/tags?per_page=1", headers=headers
    )
    if response.status_code == 200:
        data = response.json()
        if data:
            return data[0]["name"].lstrip("v")

    raise ValueError(f"Unable to determine latest version for {org}/{repo}")


def _stream_download(
    download_url: str, target_path: str, headers: dict[str, str] = REQUEST_HEADERS
) -> None:
    response = requests.get(download_url, stream=True, headers=headers)

    response.raise_for_status()

    total_size = int(response.headers.get("content-length", 0))
    progress_bar = tqdm(total=total_size, unit="iB", unit_scale=True)
    content = bytes()

    for data in response.iter_content(None, decode_unicode=True):
        progress_bar.update(len(data))
        content += data
    progress_bar.close()

    with zipfile.ZipFile(BytesIO(content)) as zf:
        zf.extractall(target_path)


def _maybe_retrieve_github_auth() -> dict[str, str]:
    """Returns appropriate github authorization headers.

    Otherwise returns an empty dict if no auth token is present.
    """
    token = os.getenv("GITHUB_TOKEN")
    if token is not None:
        auth = b64encode(token.encode()).decode()
        return {"Authorization": f"Basic {auth}"}
    return {}


def _get_download_url_from_tag(org: str, repo: str, version: str, headers: dict) -> str:
    response = requests.get(
        f"https://api.github.com/repos/{org}/{repo}/tags?per_page=100", headers=headers
    )
    response.raise_for_status()

    data = response.json()
    if not data:
        raise ValueError("Github repository has no tags set")

    available_versions = []
    for tag in data:
        tag_version = tag["name"].lstrip("v")
        available_versions.append(tag_version)
        if tag_version == version:
            return tag["zipball_url"]

    # If we've gone through all tags without finding a match, raise an error
    raise ValueError(
        f"Invalid version '{version}' for this package. Available versions are:\n"
        + ", ".join(available_versions)
    )


def _pip_installs(package_ids: list[str], base_install_path: Path, quiet: bool = False):
    logger.info(f"Installing {len(package_ids)} pip packages...")
    cmd = []
    cmd = ["uv", "pip", "install", *package_ids, "--target", str(base_install_path)]

    capture_output = quiet
    try:
        subprocess.run(cmd, capture_output=capture_output, check=True)
    except FileNotFoundError as e:
        logger.info(
            f"Stack trace:\n{''.join(traceback.format_exception(type(e), e, e.__traceback__))}"
        )
        sys.exit(1)

    _write_new_dependencies(package_ids, DependencyType.PIP)



import os
import subprocess
from pathlib import Path

from tests.conftest import (
    ANVIL1_KEYSTORE_NAME,
    ANVIL1_KEYSTORE_PASSWORD,
    ANVIL1_PRIVATE_KEY,
)


# --------------------------------------------------------------
#                         WITHOUT ANVIL
# --------------------------------------------------------------
def test_run_help(mox_path, complex_temp_path):
    current_dir = Path.cwd()
    try:
        os.chdir(complex_temp_path)
        result = subprocess.run(
            [mox_path, "run", "-h"], check=True, capture_output=True, text=True
        )
    finally:
        os.chdir(current_dir)
    assert "Moccasin CLI run" in result.stdout


def test_run_default(mox_path, complex_temp_path):
    current_dir = Path.cwd()
    try:
        os.chdir(complex_temp_path)
        result = subprocess.run(
            [mox_path, "run", "deploy"], check=True, capture_output=True, text=True
        )
    finally:
        os.chdir(current_dir)
    assert "Ending count:  1" in result.stdout


def test_multiple_manifest_returns_the_same_or_different(mox_path, complex_temp_path):
    current_dir = Path.cwd()
    os.chdir(complex_temp_path)
    try:
        result = subprocess.run(
            [mox_path, "run", "quad_manifest"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    print_statements = result.stdout.split("\n")
    assert print_statements[0] == print_statements[1] == print_statements[2]
    assert print_statements[3] != print_statements[0]
    assert_broadcast_count(print_statements, 0)


# ------------------------------------------------------------------
#                           WITH ANVIL
# ------------------------------------------------------------------
def test_run_with_network(mox_path, complex_temp_path, anvil):
    current_dir = Path.cwd()
    os.chdir(complex_temp_path)
    try:
        result = subprocess.run(
            [
                mox_path,
                "run",
                "deploy",
                "--network",
                "anvil",
                "--private-key",
                ANVIL1_PRIVATE_KEY,
            ],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert "Ending count:  1" in result.stdout
    assert "tx broadcasted" in result.stdout
    assert result.returncode == 0


def test_run_with_keystore_account(mox_path, complex_temp_path, anvil):
    current_dir = Path.cwd()
    os.chdir(complex_temp_path)
    try:
        result = subprocess.run(
            [
                mox_path,
                "run",
                "deploy",
                "--network",
                "anvil",
                "--account",
                ANVIL1_KEYSTORE_NAME,
                "--password",
                ANVIL1_KEYSTORE_PASSWORD,
            ],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert "Ending count:  1" in result.stdout
    assert "tx broadcasted" in result.stdout
    assert result.returncode == 0


def test_run_fork_should_not_send_transactions(
    mox_path,
    complex_temp_path,
    complex_project_config,
    set_fake_chain_rpc,
    anvil_two_no_state,
    anvil_keystore,
):
    current_dir = Path.cwd()
    os.chdir(complex_temp_path)
    try:
        result = subprocess.run(
            [mox_path, "run", "deploy", "--fork", "--network", "anvil-fork"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert "Ending count:  1" in result.stdout
    assert "tx broadcasted" not in result.stdout
    assert result.returncode == 0


def test_multiple_manifest_returns_the_same_or_different_on_real_network(
    mox_path, complex_temp_path, anvil
):
    current_dir = Path.cwd()
    os.chdir(complex_temp_path)
    try:
        result = subprocess.run(
            [mox_path, "run", "quad_manifest", "--network", "anvil"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    print_statements = result.stdout.split("\n")
    assert print_statements[0] == print_statements[1] == print_statements[2]
    assert print_statements[3] != print_statements[0]
    assert_broadcast_count(print_statements, 1)


def test_network_should_prompt_on_live(
    mox_path, complex_temp_path, set_fake_chain_rpc, anvil
):
    current_dir = Path.cwd()
    os.chdir(complex_temp_path)
    try:
        result = subprocess.run(
            [mox_path, "run", "deploy", "--network", "anvil-live"],
            input="y\n",
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert "Ending count:  1" in result.stdout
    assert "tx broadcasted" in result.stdout
    assert "Are you sure you wish to continue?" in result.stdout
    assert result.returncode == 0


def test_network_operation_cancelled_on_no_input(
    mox_path, complex_temp_path, set_fake_chain_rpc, anvil
):
    current_dir = Path.cwd()
    os.chdir(complex_temp_path)
    try:
        result = subprocess.run(
            [mox_path, "run", "deploy", "--network", "anvil-live"],
            input="\n",
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert "Are you sure you wish to continue?" in result.stdout
    assert "Operation cancelled." in result.stderr
    assert result.returncode == 0


def test_prompt_live_on_non_test_networks(
    mox_path, complex_temp_path, complex_project_config, anvil
):
    current_dir = Path.cwd()
    try:
        os.chdir(complex_temp_path)
        result = subprocess.run(
            [mox_path, "run", "deploy", "--network", "anvil-live"],
            input="\n",
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert (
        "The transactions run on this will actually be broadcast/transmitted, spending gas associated with your account. Are you sure you wish to continue?"
        in result.stdout
    )


# ------------------------------------------------------------------
#                            HELPERS
# ------------------------------------------------------------------
def assert_broadcast_count(print_statements: list, count: int):
    broadcast_count = 0
    for statement in print_statements:
        if "tx broadcasted" in statement:
            broadcast_count += 1
    assert broadcast_count == count



import os
from pathlib import Path

# File Names
README_PATH = "README.md"
COUNTER_CONTRACT = "Counter.vy"

# Default Project Values
BUILD_FOLDER = "out"
CONTRACTS_FOLDER = "src"
SCRIPT_FOLDER = "script"
DEPENDENCIES_FOLDER = "lib"
PYEVM = "pyevm"
ERAVM = "eravm"
SAVE_TO_DB = "save_to_db"
DEFAULT_NETWORK = PYEVM
DB_PATH_LOCAL_DEFAULT = ":memory:"
DB_PATH_LIVE_DEFAULT = ".deployments.db"

# Project Config Keys
SAVE_ABI_PATH = "save_abi_path"

# Tests folder is always tests
TESTS_FOLDER = "tests"

# Installation Variables
REQUEST_HEADERS = {"User-Agent": "Moccasin"}
PACKAGE_VERSION_FILE = "versions.toml"
PYPI = "pypi"
GITHUB = "github"

# Complex Vars
DEFAULT_PROJECT_FOLDERS = [
    CONTRACTS_FOLDER,
    TESTS_FOLDER,
    SCRIPT_FOLDER,
    DEPENDENCIES_FOLDER + "/" + GITHUB,
    DEPENDENCIES_FOLDER + "/" + PYPI,
]

# Define default values for PYEVM and ERAVM
LOCAL_NETWORK_DEFAULTS = {
    PYEVM: {
        "is_zksync": False,
        "prompt_live": False,
        SAVE_TO_DB: False,
        "live_or_staging": False,
    },
    ERAVM: {
        "is_zksync": True,
        "prompt_live": False,
        SAVE_TO_DB: False,
        "live_or_staging": False,
    },
}

FORK_NETWORK_DEFAULTS = {
    "is_zksync": False,
    "prompt_live": False,
    SAVE_TO_DB: False,
    "live_or_staging": False,
}

# Configurable Vars
MOCCASIN_DEFAULT_FOLDER = Path(
    os.getenv("MOCCASIN_DEFAULT_FOLDER", Path.home().joinpath(".moccasin/"))
)
MOCCASIN_DEFAULT_FOLDER.mkdir(parents=True, exist_ok=True)
MOCCASIN_KEYSTORE_PATH = Path(
    os.getenv("MOCCASIN_KEYSTORE_PATH", MOCCASIN_DEFAULT_FOLDER.joinpath("keystores/"))
)
MOCCASIN_KEYSTORE_PATH.mkdir(parents=True, exist_ok=True)
CONFIG_NAME = "moccasin.toml"

# Default Network Vars
FOUNDRTY_KEYSTORES_PATH = Path.home().joinpath(".foundry/keystores")
DOT_ENV_FILE = ".env"
DOT_ENV_KEY = "dot_env"
KEYSTORES_PATH_KEY = "keystores_path"
CONSOLE_HISTORY_FILE = "moccasin_history"
DEFAULT_API_KEY_ENV_VAR = "EXPLORER_API_KEY"
RESTRICTED_VALUES_FOR_LOCAL_NETWORK = [
    "url",
    "chain_id",
    "is_fork",
    "explorer_uri",
    "exploer_api_key",
]

# Testing Vars
ERA_DEFAULT_PRIVATE_KEY = (
    "0x3d3cbc973389cb26f657686445bcc75662b415b656078503592ac8c1abb8810e"
)
DEFAULT_ANVIL_PRIVATE_KEY = (
    "0xac0974bec39a17e36ba4a6b4d238ff944bacb478cbed5efcae784d7bf4f2ff80"
)
DEFAULT_ANVIL_SENDER = "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
MOCCASIN_GITHUB = "https://github.com/cyfrin/moccasin"
STARTING_BOA_BALANCE = 1000000000000000000000  # 1,000 Ether

# Database vars
GET_CONTRACT_SQL = "SELECT {} FROM deployments {}ORDER BY broadcast_ts DESC {}"
SQL_WHERE = "WHERE "
SQL_CONTRACT_NAME = "contract_name = ? "
SQL_AND = "AND "
SQL_CHAIN_ID = "json_extract(tx_dict, '$.chainId') = ? "
SQL_LIMIT = "LIMIT ? "

# Networking defaults
DEFAULT_NETWORKS_BY_NAME: dict[str, dict] = {
    "mainnet": {
        "explorer": "https://eth.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
        "chain_id": 1,
    },
    "sepolia": {
        "explorer": "https://eth-sepolia.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 11155111,
    },
    "goerli": {
        "explorer": "https://eth-goerli.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": "0x5BA1e12693Dc8F9c48aAD8770482f4739bEeD696",
        "chain_id": 5,
    },
    "arbitrum": {
        "explorer": "https://arbitrum.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": "0x5B5CFE992AdAC0C9D48E05854B2d91C73a003858",
        "chain_id": 42161,
    },
    "optimism-main": {
        "explorer": "https://optimism.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": "0x2DC0E2aa608532Da689e89e237dF582B783E552C",
        "chain_id": 10,
    },
    "optimism-test": {
        "explorer": "https://optimism-sepolia.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": "0x2DC0E2aa608532Da689e89e237dF582B783E552C",
        "chain_id": 420,
    },
    "polygon-main": {
        "explorer": "https://polygon.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": "0xc8E51042792d7405184DfCa245F2d27B94D013b6",
        "chain_id": 137,
    },
    "gnosis-main": {
        "explorer": "https://gnosis.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 100,
    },
    "gnosis-test": {
        "explorer": "https://gnosis-chiado.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 10200,
    },
    "base-main": {
        "explorer": "https://base.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 8453,
    },
    "syscoin": {
        "explorer": "https://explorer.syscoin.org/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 57,
    },
    "ethereumclassic": {
        "explorer": "https://etc.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 61,
    },
    "nova-network": {
        "explorer": "https://explorer.novanetwork.io/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 87,
    },
    "velas": {
        "explorer": "https://evmexplorer.velas.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 106,
    },
    "thundercore": {
        "explorer": "https://explorer-mainnet.thundercore.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 108,
    },
    "fuse": {
        "explorer": "https://explorer.fuse.io/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 122,
    },
    "heco": {"explorer": None, "multicall2": None, "chain_id": 128},
    "shimmer_evm": {
        "explorer": "https://explorer.evm.shimmer.network/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 148,
    },
    "manta": {
        "explorer": "https://pacific-explorer.manta.network/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 169,
    },
    "energyweb": {
        "explorer": "https://explorer.energyweb.org/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 246,
    },
    "oasys": {
        "explorer": "https://explorer.oasys.games/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 248,
    },
    "omax": {"explorer": "https://omaxscan.com/", "multicall2": None, "chain_id": 311},
    "filecoin": {
        "explorer": "https://filecoin.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 314,
    },
    "kucoin": {"explorer": "https://scan.kcc.io/", "multicall2": None, "chain_id": 321},
    "zksync-era": {
        "explorer": "https://zksync2-mainnet-explorer.zksync.io",
        "explorer_type": "zksyncexplorer",
        "multicall2": None,
        "chain_id": 324,
    },
    "sepolia-zksync-era": {
        "explorer": "https://explorer.sepolia.era.zksync.dev",
        "explorer_type": "zksyncexplorer",
        "multicall2": None,
        "chain_id": 300,
    },
    "shiden": {
        "explorer": "https://shiden.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 336,
    },
    "rollux": {
        "explorer": "https://explorer.rollux.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 570,
    },
    "astar": {
        "explorer": "https://astar.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 592,
    },
    "callisto": {
        "explorer": "https://explorer.callisto.network/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 820,
    },
    "lyra-chain": {
        "explorer": "https://explorer.lyra.finance/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 957,
    },
    "bifrost": {
        "explorer": "https://explorer.mainnet.bifrostnetwork.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 996,
    },
    "metis": {
        "explorer": "https://andromeda-explorer.metis.io/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 1088,
    },
    "polygon-zkevm": {
        "explorer": "https://zkevm.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 1101,
    },
    "core": {"explorer": None, "multicall2": None, "chain_id": 1116},
    "lisk": {
        "explorer": "https://blockscout.lisk.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 1135,
    },
    "reya-network": {
        "explorer": "https://explorer.reya.network/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 1729,
    },
    "onus": {
        "explorer": "https://explorer.onuschain.io/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 1975,
    },
    "hubblenet": {"explorer": None, "multicall2": None, "chain_id": 1992},
    "sanko": {
        "explorer": "https://explorer.sanko.xyz/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 1996,
    },
    "dogechain": {
        "explorer": "https://explorer.dogechain.dog/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 2000,
    },
    "milkomeda": {
        "explorer": "https://explorer-mainnet-cardano-evm.c1.milkomeda.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 2001,
    },
    "kava": {
        "explorer": "https://testnet.kavascan.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 2222,
    },
    "mantle": {
        "explorer": "https://explorer.mantle.xyz/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 5000,
    },
    "zetachain": {
        "explorer": "https://zetachain.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 7000,
    },
    "celo": {
        "explorer": "https://explorer.celo.org/mainnet/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 42220,
    },
    "oasis": {
        "explorer": "https://explorer.emerald.oasis.dev/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 42262,
    },
    "linea": {
        "explorer": "https://explorer.linea.build/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 59144,
    },
    "blast": {
        "explorer": "https://blast.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 81457,
    },
    "taiko": {
        "explorer": "https://blockscoutapi.mainnet.taiko.xyz/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 167000,
    },
    "scroll": {
        "explorer": "https://blockscout.scroll.io/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 534352,
    },
    "zora": {
        "explorer": "https://explorer.zora.energy/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 7777777,
    },
    "neon": {
        "explorer": "https://neon.blockscout.com/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 245022934,
    },
    "aurora": {
        "explorer": "https://explorer.mainnet.aurora.dev/",
        "explorer_type": "blockscout",
        "multicall2": None,
        "chain_id": 1313161554,
    },
}

DEFAULT_NETWORKS_BY_CHAIN_ID: dict[int, dict] = {
    network["chain_id"]: {**network, "name": name}
    for name, network in DEFAULT_NETWORKS_BY_NAME.items()
}



import os
import subprocess
from pathlib import Path

from moccasin.commands.install import GITHUB, PYPI
from moccasin.config import Config
from moccasin.constants.vars import DEPENDENCIES_FOLDER
from tests.conftest import (
    comment_content,
    github_package_name,
    new_version,
    org_name,
    patrick_package_name,
    pip_package_name,
    version,
)


def test_install_without_parameters_installs_packages_in_toml(
    installation_temp_path, installation_cleanup_dependencies, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(installation_temp_path)
        result = subprocess.run(
            [mox_path, "install"], check=True, capture_output=True, text=True
        )
    finally:
        os.chdir(current_dir)
    assert f"+ {pip_package_name}=={version}" in result.stderr
    assert (
        Path(installation_temp_path)
        .joinpath(f"{DEPENDENCIES_FOLDER}/{PYPI}/{pip_package_name}")
        .exists()
    )


def test_double_install_snekmate(
    installation_temp_path, installation_cleanup_dependencies, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(installation_temp_path)
        result_one = subprocess.run(
            [mox_path, "install", github_package_name],
            check=True,
            capture_output=True,
            text=True,
        )
        result_two = subprocess.run(
            [mox_path, "install", github_package_name],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert result_one.returncode == 0
    assert result_two.returncode == 0
    assert (
        Path(installation_temp_path)
        .joinpath(f"{DEPENDENCIES_FOLDER}/{GITHUB}/{org_name}")
        .exists()
    )


def test_write_to_config_after_install(
    installation_temp_path, installation_cleanup_dependencies, mox_path
):
    project_root: Path = Config.find_project_root(Path(installation_temp_path))
    config = Config(project_root)
    starting_dependencies = config.dependencies
    assert github_package_name not in config.dependencies

    # Arrange
    current_dir = Path.cwd()
    # Act
    try:
        os.chdir(installation_temp_path)
        subprocess.run(
            [mox_path, "install", github_package_name],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    # Assert
    project_root: Path = Config.find_project_root(Path(installation_temp_path))
    config = Config(project_root)
    assert github_package_name in config.dependencies
    for dep in starting_dependencies:
        assert dep in config.dependencies
    assert comment_content in config.read_configs_preserve_comments().as_string()


def test_can_install_with_version(
    installation_temp_path, installation_cleanup_dependencies, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(installation_temp_path)
        result = subprocess.run(
            [mox_path, "install", f"{github_package_name}@{version}"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert f"Installed {github_package_name}" in result.stderr
    project_root: Path = Config.find_project_root(Path(installation_temp_path))
    config = Config(project_root)
    assert f"{github_package_name}@{version}" in config.dependencies
    assert (
        Path(installation_temp_path)
        .joinpath(f"{DEPENDENCIES_FOLDER}/{GITHUB}/{org_name}")
        .exists()
    )


def test_can_change_versions(
    installation_temp_path, installation_cleanup_dependencies, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(installation_temp_path)
        result_one = subprocess.run(
            [mox_path, "install", f"{github_package_name}@{version}"],
            check=True,
            capture_output=True,
            text=True,
        )
        result_two = subprocess.run(
            [mox_path, "install", f"{github_package_name}@{new_version}"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert f"Installed {github_package_name}" in result_one.stderr
    assert f"Updating {github_package_name}" in result_two.stderr
    project_root: Path = Config.find_project_root(Path(installation_temp_path))
    config = Config(project_root)
    assert f"{github_package_name}@{new_version}" in config.dependencies
    assert (
        Path(installation_temp_path)
        .joinpath(f"{DEPENDENCIES_FOLDER}/{GITHUB}/{org_name}")
        .exists()
    )


def test_can_compile_with_github_search_path(
    installation_temp_path, installation_cleanup_dependencies, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(installation_temp_path)
        result_install = subprocess.run(
            [mox_path, "install", patrick_package_name],
            check=True,
            capture_output=True,
            text=True,
        )
        result_compile = subprocess.run(
            [mox_path, "compile"], check=True, capture_output=True, text=True
        )
    finally:
        os.chdir(current_dir)
    assert "Done compiling project!" in result_compile.stderr
    assert result_install.returncode == 0
    assert result_compile.returncode == 0


def test_no_moccasin_toml_saves_dependencies_to_pyproject(
    mox_path, no_config_temp_path, no_config_config
):
    current_dir = Path.cwd()
    try:
        os.chdir(no_config_temp_path)
        result_install = subprocess.run(
            [mox_path, "install", patrick_package_name],
            check=True,
            capture_output=True,
            text=True,
        )
        result_compile = subprocess.run(
            [mox_path, "compile"], check=True, capture_output=True, text=True
        )
    finally:
        os.chdir(current_dir)
    assert no_config_config.read_pyproject_config() == {
        "project": {
            "dot_env": ".hello",
            "src": "contracts",
            "dependencies": ["patrickalphac/test_repo"],
        }
    }

    with open(no_config_temp_path.joinpath("pyproject.toml")) as f:
        assert "patrickalphac/test_repo" in f.read()
    assert "Done compiling project!" in result_compile.stderr
    assert result_install.returncode == 0
    assert result_compile.returncode == 0

    project_root: Path = Config.find_project_root(Path(no_config_temp_path))
    config = Config(project_root)
    assert patrick_package_name in config.dependencies



import os
import shutil
import tempfile
import tomllib
import warnings
from dataclasses import dataclass, field
from pathlib import Path
from typing import TYPE_CHECKING, Any, Iterator, Tuple, Union, cast

import boa
import tomlkit
from boa.contracts.abi.abi_contract import ABIContract, ABIContractFactory
from boa.contracts.vyper.vyper_contract import (
    VyperContract,
    VyperDeployer,
    build_abi_output,
)
from boa.deployments import (
    Deployment,
    DeploymentsDB,
    get_deployments_db,
    set_deployments_db,
)
from boa.environment import Env
from boa.util.abi import Address
from boa.verifiers import get_verification_bundle
from boa_zksync import set_zksync_env, set_zksync_fork, set_zksync_test_env
from boa_zksync.contract import ZksyncContract
from boa_zksync.deployer import ZksyncDeployer
from dotenv import load_dotenv
from eth_utils import to_hex
from tomlkit.items import Table

from moccasin.constants.vars import (
    BUILD_FOLDER,
    CONFIG_NAME,
    CONTRACTS_FOLDER,
    DB_PATH_LIVE_DEFAULT,
    DB_PATH_LOCAL_DEFAULT,
    DEFAULT_NETWORK,
    DEPENDENCIES_FOLDER,
    DOT_ENV_FILE,
    DOT_ENV_KEY,
    ERAVM,
    FORK_NETWORK_DEFAULTS,
    GET_CONTRACT_SQL,
    LOCAL_NETWORK_DEFAULTS,
    PYEVM,
    RESTRICTED_VALUES_FOR_LOCAL_NETWORK,
    SAVE_ABI_PATH,
    SAVE_TO_DB,
    SCRIPT_FOLDER,
    SQL_AND,
    SQL_CHAIN_ID,
    SQL_CONTRACT_NAME,
    SQL_LIMIT,
    SQL_WHERE,
    TESTS_FOLDER,
)
from moccasin.logging import logger
from moccasin.moccasin_account import MoccasinAccount
from moccasin.named_contract import NamedContract

if TYPE_CHECKING:
    from boa.network import NetworkEnv
    from boa.verifiers import Blockscout, VerificationResult
    from boa_zksync import ZksyncEnv
    from boa_zksync.verifiers import ZksyncExplorer

_AnyEnv = Union["NetworkEnv", "Env", "ZksyncEnv"]
VERIFIERS = Union["Blockscout", "ZksyncExplorer"]


@dataclass
class Network:
    name: str
    url: str | None = None
    chain_id: int | None = None
    is_fork: bool = False
    is_zksync: bool = False
    default_account_name: str | None = None
    unsafe_password_file: Path | None = None
    save_abi_path: str | None = None
    explorer_uri: str | None = None
    explorer_api_key: str | None = None
    explorer_type: str | None = None
    contracts: dict[str, NamedContract] = field(default_factory=dict)
    prompt_live: bool = True
    save_to_db: bool = True
    live_or_staging: bool = True
    db_path: str | Path = DB_PATH_LOCAL_DEFAULT
    extra_data: dict[str, Any] = field(default_factory=dict)
    _network_env: _AnyEnv | None = None

    def _set_boa_env(self) -> _AnyEnv:
        """Sets the boa.env to the current network, this additionally sets up the database."""
        # perf: save time on imports in the (common) case where
        # we just import config for its utils but don't actually need
        # to switch networks
        from boa.network import EthereumRPC, NetworkEnv

        # 1. Check for forking, and set (You cannot fork from a NetworkEnv, only a "new" Env!)
        if self.is_fork:
            if self.is_zksync:
                set_zksync_fork(url=self.url)
            else:
                boa.fork(self.url)

        # 2. Non-forked local networks
        elif self.name == PYEVM:
            env = Env()
            boa.set_env(env)
        elif self.name == ERAVM:
            set_zksync_test_env()

        # 3. Finally, "true" networks
        elif self.is_zksync:
            if self.explorer_type != "zksyncexplorer":
                logger.warning(
                    "Explorer type is not zksyncexplorer, as of today, only the zksync explorer is supported with zksync."
                )
            set_zksync_env(self.url, explorer_url=self.explorer_uri, nickname=self.name)
        else:
            env = NetworkEnv(EthereumRPC(self.url), nickname=self.name)
            boa.set_env(env)

        # 4. Quick sanity check on chain_id
        if not self.is_local_or_forked_network():
            expected_chain_id = boa.env.get_chain_id()  # type: ignore
            if self.chain_id and self.chain_id != expected_chain_id:
                raise ValueError(
                    f"Chain ID mismatch. Expected {expected_chain_id}, but user placed {self.chain_id} in their {CONFIG_NAME}."
                )
            self.chain_id = expected_chain_id if not self.chain_id else self.chain_id

            # 5. Set the explorer type and uri if not set by default ones
            if self.chain_id is not None and (
                self.explorer_uri is None or self.explorer_uri is None
            ):
                # Review: Is this down here even good?
                from moccasin.constants.vars import DEFAULT_NETWORKS_BY_CHAIN_ID

                default_network_info = DEFAULT_NETWORKS_BY_CHAIN_ID.get(
                    self.chain_id, {}
                )
                self.explorer_uri = (
                    self.explorer_uri
                    if self.explorer_uri is not None
                    else default_network_info.get("explorer", None)
                )
                self.explorer_type = (
                    self.explorer_type
                    if self.explorer_type is not None
                    else default_network_info.get("explorer_type", None)
                )

        boa.env.nickname = self.name
        return boa.env

    def moccasin_verify(
        self, contract: VyperContract | ZksyncContract
    ) -> "VerificationResult":
        """Verifies a contract using your moccasin.toml config."""
        verifier_class = self.get_verifier_class()
        verifier_instance = verifier_class(self.explorer_uri, self.explorer_api_key)
        if self.is_zksync:
            import boa_zksync

            return boa_zksync.verify(contract, verifier_instance)
        return boa.verify(contract, verifier_instance)

    def is_matching_boa(self) -> bool:
        """Returns True if the current network is the active network in boa.
        This is a good way to test if you've overriden boa as the "active" network.
        """
        return boa.env.nickname == self.name

    def get_verifier_class(self) -> Any:
        if self.explorer_type is None:
            if self.explorer_uri is not None:
                if "blockscout" in self.explorer_uri:
                    self.explorer_type = "blockscout"
                elif "zksync" in self.explorer_uri:
                    self.explorer_type = "zksyncexplorer"

        if self.explorer_type is None:
            raise ValueError(
                f"No explorer type found. Please set the explorer_type in your {CONFIG_NAME}."
            )
        if self.explorer_uri is None:
            raise ValueError(
                f"No explorer_uri found. Please set the `explorer_uri` in your {CONFIG_NAME}."
            )

        verifier_name = self._to_verifier_name(self.explorer_type)
        from importlib import import_module

        module = import_module("moccasin.supported_verifiers")
        return getattr(module, verifier_name)

    def _to_verifier_name(self, verifier_string: str) -> str:
        if verifier_string.lower().strip() == "blockscout":
            return "Blockscout"
        if verifier_string.lower().strip() == "zksyncexplorer":
            return "ZksyncExplorer"
        raise ValueError(
            f"Verifier {verifier_string} is not supported. Please use 'blockscout' or 'zksyncexplorer'."
        )

    def get_default_account(self) -> MoccasinAccount | Any:
        """Returns an 'account-like' object."""
        if hasattr(boa.env, "_accounts"):
            if boa.env.eoa is not None:
                return boa.env._accounts[boa.env.eoa]
        else:
            if boa.env.eoa is not None:
                return MoccasinAccount(address=boa.env.eoa, ignore_warning=True)
        if (
            self.default_account_name is not None
            and self.unsafe_password_file is not None
        ):
            return MoccasinAccount(
                keystore_path_or_account_name=self.default_account_name,
                password_file_path=self.unsafe_password_file,
            )
        return None

    def set_kwargs(self, **kwargs):
        for key, value in kwargs.items():
            if value is not None:
                setattr(self, key, value)

    def _set_boa_db(self) -> None:
        db: DeploymentsDB
        if self.save_to_db:
            db = DeploymentsDB(path=self.db_path)
        else:
            db = DeploymentsDB(path=DB_PATH_LOCAL_DEFAULT)
        set_deployments_db(db)

    def create_and_set_or_set_boa_env(self, **kwargs) -> _AnyEnv:
        self.set_kwargs(**kwargs)
        self._set_boa_env()
        self._network_env = boa.env
        return self._network_env

    def _generate_sql_from_args(
        self,
        contract_name: str | None = None,
        chain_id: int | str | None = None,
        limit: int | None = None,
        db: DeploymentsDB | None = None,
    ) -> tuple[str, tuple]:
        if db is None:
            db = get_deployments_db()

        where_clauses = []
        params: list[str | int] = []

        field_names = db._get_fieldnames_str()

        if contract_name is not None:
            where_clauses.append(SQL_CONTRACT_NAME)
            params.append(contract_name)

        # Add chain_id condition if provided
        if chain_id is not None:
            where_clause = SQL_CHAIN_ID
            if where_clauses:
                where_clause = SQL_AND + where_clause
            where_clauses.append(where_clause)
            params.append(str(chain_id))

        where_part = ""
        if where_clauses:
            where_part = SQL_WHERE + "".join(where_clauses)

        # Add LIMIT if provided
        limit_part = ""
        if limit is not None:
            limit_part = SQL_LIMIT
            params.append(int(limit))

        sql_query = GET_CONTRACT_SQL.format(field_names, where_part, limit_part)
        return sql_query, tuple(params)

    def _fetch_deployments_from_db(
        self,
        contract_name: str | None = None,
        chain_id: int | str | None = None,
        limit: int | None = None,
        db: DeploymentsDB | None = None,
    ) -> Iterator[Deployment]:
        if db is None:
            db = get_deployments_db()
        chain_id = to_hex(chain_id) if chain_id is not None else None
        if not isinstance(limit, int) and not isinstance(limit, type(None)):
            raise ValueError(f"Limit must be an integer, not {type(limit)}.")
        final_sql, params = self._generate_sql_from_args(
            contract_name=contract_name, chain_id=chain_id, limit=limit, db=db
        )
        return db._get_deployments_from_sql(final_sql, params)

    def _get_deployments_iterator(
        self,
        contract_name: str | None = None,
        chain_id: int | str | None = None,
        limit: int | None = None,
        config_or_db_path: Union["Config", Path, str, None] = None,
    ) -> Iterator[Deployment]:
        """Get deployments from the database without an initialized config."""
        db_path = None
        if isinstance(config_or_db_path, Config):
            db_path = config_or_db_path._toml_data.get("db_path", ".deployments.db")
        elif isinstance(db_path, str):
            db_path = Path(db_path)
        elif isinstance(config_or_db_path, Path):
            db_path = config_or_db_path

        if not db_path:
            db = get_deployments_db()
        else:
            db = DeploymentsDB(db_path)
        return self._fetch_deployments_from_db(
            contract_name=contract_name, chain_id=chain_id, limit=limit, db=db
        )

    def get_deployments_unchecked(
        self,
        contract_name: str | None = None,
        limit: int | None = None,
        chain_id: int | str | None = None,
        config_or_db_path: Union["Config", Path, str, None] = None,
    ) -> list[Deployment]:
        deployments_iter = self._get_deployments_iterator(
            contract_name=contract_name,
            chain_id=chain_id,
            limit=limit,
            config_or_db_path=config_or_db_path,
        )
        return list(deployments_iter)

    def get_deployments_checked(
        self,
        contract_name: str | None = None,
        limit: int | None = None,
        chain_id: int | str | None = None,
        config_or_db_path: Union["Config", Path, str, None] = None,
    ) -> list[Deployment]:
        deployments_iter = self._get_deployments_iterator(
            contract_name=contract_name,
            chain_id=chain_id,
            limit=limit,
            config_or_db_path=config_or_db_path,
        )

        config = config_or_db_path
        if not isinstance(config_or_db_path, Config):
            config = get_config()
        config = cast(Config, config)
        deployments_list = []
        for deployment in deployments_iter:
            if self.has_matching_integrity(deployment, contract_name, config=config):
                deployments_list.append(deployment)
        return deployments_list

    def has_matching_integrity(
        self,
        deployment: Deployment,
        contract_name: str | None,
        config: Union["Config", None] = None,
    ) -> bool:
        if config is None:
            config = get_config()
        if contract_name is None:
            raise ValueError("contract_name cannot be None.")
        vyper_deployer = self.get_deployer_from_contract_name(config, contract_name)
        # REVIEW: If boa throws a warning, maybe that's all we need instead of checking the integrity
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="casted bytecode does not match compiled bytecode at*"
            )
            contract = vyper_deployer.at(deployment.contract_address)
        verification_bundle = get_verification_bundle(contract)
        expected_integrity = verification_bundle["integrity"]
        actual_integrity = deployment.source_code["integrity"]
        if expected_integrity != actual_integrity:
            return False
        return True

    def get_deployer_from_contract_name(
        self, config: "Config", contract_name: str
    ) -> VyperDeployer | ZksyncDeployer:
        contract_path = config.find_contract(contract_name)
        return boa.load_partial(str(contract_path.absolute()))

    def get_latest_deployment_unchecked(
        self, contract_name: str | None = None, chain_id: int | str | None = None
    ) -> Deployment | None:
        deployments = self.get_deployments_unchecked(
            contract_name=contract_name, chain_id=chain_id, limit=1
        )
        if len(deployments) > 0:
            return deployments[0]
        return None

    def get_latest_contract_unchecked(
        self, contract_name: str | None = None, chain_id: int | str | None = None
    ) -> ABIContract | None:
        deployment = self.get_latest_deployment_unchecked(
            contract_name=contract_name, chain_id=chain_id
        )
        if deployment is not None:
            return self.convert_deployment_to_contract(deployment)
        return None

    def get_latest_deployment_checked(
        self, contract_name: str | None = None, chain_id: int | str | None = None
    ) -> Deployment | None:
        deployments = self.get_deployments_checked(
            contract_name=contract_name, chain_id=chain_id, limit=1
        )
        if len(deployments) > 0:
            return deployments[0]
        return None

    def get_latest_contract_checked(
        self, contract_name: str | None = None, chain_id: int | str | None = None
    ) -> ABIContract | None:
        deployment = self.get_latest_deployment_checked(
            contract_name=contract_name, chain_id=chain_id
        )
        if deployment is not None:
            return self.convert_deployment_to_contract(deployment)
        return None

    def manifest_contract(
        self,
        contract_name: str,
        force_deploy: bool = False,
        address: str | None = None,
        checked: bool = False,
    ) -> VyperContract | ZksyncContract | ABIContract:
        """A wrapper around get_or_deploy_contract that is more explicit about the contract being deployed."""
        return self.get_or_deploy_contract(
            contract_name=contract_name,
            force_deploy=force_deploy,
            address=address,
            checked=checked,
        )

    def instantiate_contract(
        self, *args, **kwargs
    ) -> VyperContract | ZksyncContract | ABIContract:
        """An alias for get_or_deploy_contract."""
        return self.get_or_deploy_contract(*args, **kwargs)

    def get_or_deploy_contract(
        self,
        contract_name: str,
        force_deploy: bool = False,
        abi: str
        | Path
        | list
        | VyperDeployer
        | VyperContract
        | ZksyncContract
        | ZksyncDeployer
        | ABIContractFactory
        | ABIContract
        | None = None,
        abi_from_explorer: bool | None = None,
        deployer_script: str | Path | None = None,
        address: str | None = None,
        checked: bool = False,
    ) -> VyperContract | ZksyncContract | ABIContract:
        """Returns or deploys a VyperContract, ZksyncContract, or ABIContract based on the name and address in the config file, or passed to this function.

        The following arguments are mutually exclusive:
            - abi
            - abi_from_explorer

        Args:
            - contract_name: the contract name or deployer for the contract.
            - force_deploy: if True, will deploy the contract even if the contract has an address in the config file.
            - abi: the ABI of the contract. Can be a list, string path to file, string of the ABI, VyperDeployer, VyperContract, ZksyncDeployer, ZksyncContract, ABIContractFactory, or ABIContract.
            - abi_from_explorer: if True, will fetch the ABI from the explorer.
            - deployer_script: If no address is given, this is the path to deploy the contract.
            - address: The address of the contract.

        Returns:
            VyperContract | ZksyncContract | ABIContract: The deployed contract instance, or a blank contract if the contract is not found.
        """
        named_contract: NamedContract = self.contracts.get(
            contract_name, NamedContract(contract_name)
        )
        if not named_contract.is_active():
            if self.contracts.get(contract_name) is not None:
                self.contracts[contract_name]._reset()

        if abi_from_explorer and abi:
            raise ValueError(
                "abi and abi_from_explorer are mutually exclusive. Please only provide one."
            )
        abi_or_deployer = abi

        # 0. Setup parameters based on defaults and the inputs to this func
        if not abi_or_deployer and not abi_from_explorer:
            abi_or_deployer = named_contract.get("abi", None)
            vyper_deployer = named_contract.get("vyper_deployer", None)
            if vyper_deployer is not None and abi_or_deployer is None:
                abi_or_deployer = vyper_deployer
        if not abi_from_explorer and not abi_or_deployer:
            abi_from_explorer = named_contract.get("abi_from_explorer", None)
        if not force_deploy:
            force_deploy = named_contract.get("force_deploy", False)
        if not deployer_script:
            deployer_script = named_contract.get("deployer_script", None)
        if not address:
            address = named_contract.get("address", None)

        # 1. Check if force_deploy is true
        if force_deploy:
            if not deployer_script:
                raise ValueError(
                    f"Contract {named_contract.contract_name} has force_deploy=True but no deployer_script specified in their {CONFIG_NAME}."
                )
            return self._deploy_named_contract(named_contract, deployer_script)

        # 2. Setup ABI based on parameters
        (abi, deployer) = self._get_abi_and_deployer_from_params(
            named_contract.contract_name, abi_or_deployer, abi_from_explorer, address
        )
        abi = abi if abi else named_contract.abi
        deployer = deployer if deployer else named_contract.vyper_deployer

        # ------------------------------------------------------------------
        #             CHECK TO SEE IF WE'VE ALREADY DEPLOYED
        # ------------------------------------------------------------------
        # 3. Happy path, check if the requested contract is what we've already deployed
        # We don't need to check the DB since we are checking the address on this network
        if (
            str(named_contract.abi) == str(abi)
            and named_contract.address == address
            and named_contract.vyper_contract is not None
        ):
            return named_contract.vyper_contract

        # 4. Happy path, we check for this contract in the DB
        if self.chain_id is not None and self.save_to_db is True:
            vyper_contract: ABIContract | None = None
            if checked:
                vyper_contract = self.get_latest_contract_checked(
                    named_contract.contract_name
                )
            else:
                vyper_contract = self.get_latest_contract_unchecked(
                    named_contract.contract_name
                )
            if vyper_contract is not None:
                return vyper_contract

        # 5. Happy path, maybe we didn't deploy the contract, but we've been given an abi and address, which works
        if abi and address:
            if deployer:
                return deployer.at(address)
            else:
                # Note, we are not putting this into the self.contracts dict, but maybe we should
                return ABIContractFactory(named_contract.contract_name, abi).at(address)

        # ------------------------------------------------------------------
        #                      WE DEPLOY AFTER HERE
        # ------------------------------------------------------------------
        # 6. Check if there is an address, if no ABI, return a blank contract at an address
        if address and not abi:
            logger.info(
                f"No abi_source or abi_path found for {named_contract.contract_name}, returning a blank contract at {address}"
            )
            # We could probably put this into _deploy_named_contract with a conditional
            blank_contract: VyperDeployer | ZksyncDeployer = boa.loads_partial("")
            vyper_contract = blank_contract.at(address)
            named_contract.update_from_deployed_contract(vyper_contract)
            self.contracts[named_contract.contract_name] = named_contract
            return vyper_contract

        # 7. If no address, deploy the contract
        if not deployer_script:
            raise ValueError(
                f"Contract {named_contract.contract_name} has no address or deployer_script specified in the {CONFIG_NAME}."
            )

        return self._deploy_named_contract(named_contract, deployer_script)

    def is_local_or_forked_network(self) -> bool:
        """Returns True if network is:
        1. pyevm
        2. eravm
        3. A fork
        """
        return self._is_local_or_forked_network(self.name, self.is_fork)

    def has_explorer(self) -> bool:
        return self.explorer_uri is not None

    def _deploy_named_contract(
        self, named_contract: NamedContract, deployer_script: str | Path
    ) -> VyperContract | ZksyncContract:
        config = get_config()
        deployed_named_contract: VyperContract | ZksyncContract = (
            named_contract._deploy(
                config.script_folder, deployer_script, update_from_deploy=True
            )
        )
        self.contracts[named_contract.contract_name] = named_contract
        return deployed_named_contract

    def _get_abi_and_deployer_from_params(
        self,
        logging_contract_name: str,
        abi: str
        | Path
        | list
        | VyperDeployer
        | VyperContract
        | ZksyncContract
        | ZksyncDeployer
        | ABIContractFactory
        | ABIContract
        | None = None,
        abi_from_explorer: bool | None = None,
        address: str | None = None,
    ) -> Tuple[Union[str, None], Union[VyperDeployer, ZksyncDeployer, None]]:
        if abi_from_explorer and not address:
            raise ValueError(
                f"Cannot get ABI from explorer without an address for contract {logging_contract_name}. Please provide an address."
            )

        abi_like = abi
        abi = None
        deployer = None

        config = get_config()
        if abi_like:
            if isinstance(abi_like, str):
                # Check if it's a file path
                if (
                    abi_like.endswith(".json")
                    or abi_like.endswith(".vy")
                    or abi_like.endswith(".vyi")
                ):
                    abi_path = config.find_contract(abi_like)
                    if abi_like.endswith(".vy"):
                        deployer = boa.load_partial(str(abi_path))
                        abi = str(build_abi_output(deployer.compiler_data))
                        return abi, deployer
                    elif abi_path.suffix == ".json":
                        abi = str(boa.load_abi(str(abi_path)).abi)
                        return abi, deployer
                    elif abi_path.suffix == ".vyi":
                        raise NotImplementedError(
                            f"Loading an ABI from Vyper Interface files is not yet supported for contract {logging_contract_name}. You can track the progress here:\nhttps://github.com/vyperlang/vyper/issues/4232"
                        )
                else:
                    contract_path = config.find_contract(abi_like)
                    deployer = boa.load_partial(str(contract_path))
                    abi = str(build_abi_output(deployer.compiler_data))
                    return abi, deployer
            if isinstance(abi_like, list):
                # If its an ABI, just take it
                return str(abi_like), deployer
            if isinstance(abi_like, VyperDeployer) or isinstance(
                abi_like, ZksyncDeployer
            ):
                return str(build_abi_output(abi_like.compiler_data)), abi_like
            if isinstance(abi_like, VyperContract) or isinstance(
                abi_like, ZksyncContract
            ):
                return abi_like.abi, abi_like.deployer
            if isinstance(abi_like, ABIContractFactory):
                return abi_like.abi, deployer
            if isinstance(abi_like, ABIContract):
                return abi_like.abi, deployer
        if abi_from_explorer:
            from moccasin.commands.explorer import boa_get_abi_from_explorer

            abi = boa_get_abi_from_explorer(str(address), quiet=True)
        return abi, deployer  # type: ignore

    def get_named_contract(self, contract_name: str) -> NamedContract | None:
        return self.contracts.get(contract_name, None)

    def get_named_contracts(self) -> dict:
        return self.contracts

    def set_boa_eoa(self, account: MoccasinAccount):
        if self.is_local_or_forked_network:  # type: ignore[truthy-function]
            boa.env.eoa = Address(account.address)
        else:
            boa.env.add_account(account, force_eoa=True)

    @property
    def alias(self) -> str:
        return self.name

    @property
    def identifier(self) -> str:
        return self.name

    @staticmethod
    def convert_deployment_to_contract(deployment: Deployment) -> ABIContract:
        contract_factory = ABIContractFactory(
            deployment.contract_name,
            deployment.abi,
            deployment.contract_name + "_" + deployment.source_code["integrity"],
        )
        return contract_factory.at(deployment.contract_address)

    @staticmethod
    def _is_local_or_forked_network(name: str, fork: bool = False) -> bool:
        return name in [PYEVM, ERAVM] or fork


class _Networks:
    _networks: dict[str, Network]
    _default_named_contracts: dict[str, NamedContract]
    _overriden_active_network: Network | None
    db_path: Path
    default_network_name: str

    def __init__(self, toml_data: dict, db_path: Path):
        self._networks = {}
        self._default_named_contracts = {}
        self._overriden_active_network = None
        self.custom_networks_counter = 0
        self.db_path = db_path
        project_data = toml_data.get("project", {})

        default_explorer_api_key = project_data.get("explorer_api_key", None)
        default_explorer_uri = project_data.get("explorer_uri", None)
        default_explorer_type = project_data.get("explorer_type", None)
        default_save_abi_path = project_data.get("save_abi_path", None)
        default_contracts = toml_data.get("networks", {}).get("contracts", {})
        self.default_network_name = project_data.get(
            "default_network_name", DEFAULT_NETWORK
        )

        self._validate_network_contracts_dict(default_contracts)
        for contract_name, contract_data in default_contracts.items():
            self._default_named_contracts[contract_name] = NamedContract(
                contract_name,
                force_deploy=contract_data.get("force_deploy", None),
                abi=contract_data.get("abi", None),
                abi_from_explorer=contract_data.get("abi_from_explorer", None),
                deployer_script=contract_data.get("deployer_script", None),
                address=contract_data.get("address", None),
            )

        toml_data = self._add_local_network_defaults(toml_data)

        for network_name, network_data in toml_data["networks"].items():
            # Check for restricted items for pyevm or eravm
            if network_name in [PYEVM, ERAVM]:
                self._validate_local_network_data(network_data, network_name)
            if network_name == "contracts":
                continue
            else:
                starting_network_contracts_dict = network_data.get("contracts", {})
                self._validate_network_contracts_dict(
                    starting_network_contracts_dict, network_name=network_name
                )
                final_network_contracts = (
                    self._generate_network_contracts_from_defaults(
                        self._default_named_contracts.copy(),
                        starting_network_contracts_dict,
                    )
                )
                if network_data.get("fork", None) is True:
                    network_data = self._add_fork_network_defaults(network_data)
                    self._validate_fork_network_defaults(network_data)
                network = Network(
                    name=network_name,
                    is_fork=network_data.get("fork", False),
                    url=network_data.get("url", None),
                    is_zksync=network_data.get("is_zksync", False),
                    chain_id=network_data.get("chain_id", None),
                    save_abi_path=network_data.get(
                        SAVE_ABI_PATH, default_save_abi_path
                    ),
                    explorer_uri=network_data.get("explorer_uri", default_explorer_uri),
                    explorer_api_key=network_data.get(
                        "explorer_api_key", default_explorer_api_key
                    ),
                    explorer_type=network_data.get(
                        "explorer_type", default_explorer_type
                    ),
                    default_account_name=network_data.get("default_account_name", None),
                    unsafe_password_file=network_data.get("unsafe_password_file", None),
                    prompt_live=network_data.get("prompt_live", True),
                    save_to_db=network_data.get(SAVE_TO_DB, True),
                    live_or_staging=network_data.get("live_or_staging", True),
                    db_path=self.db_path,
                    contracts=final_network_contracts,
                    extra_data=network_data.get("extra_data", {}),
                )
                setattr(self, network_name, network)
                self._networks[network_name] = network

    def __len__(self):
        return len(self._networks)

    def get_networks(self) -> dict[str, Network]:
        return self._networks

    def _generate_network_contracts_from_defaults(
        self, starting_default_contracts: dict, starting_network_contracts_dict: dict
    ) -> dict:
        for contract_name, contract_data in starting_network_contracts_dict.items():
            named_contract = NamedContract(
                contract_name,
                force_deploy=contract_data.get("force_deploy", None),
                abi=contract_data.get("abi", None),
                abi_from_explorer=contract_data.get("abi_from_explorer", None),
                deployer_script=contract_data.get("deployer_script", None),
                address=contract_data.get("address", None),
            )
            if self._default_named_contracts.get(contract_name, None):
                named_contract.set_defaults(
                    self._default_named_contracts[contract_name]
                )
            starting_default_contracts[contract_name] = named_contract
        return starting_default_contracts

    def get_active_network(self) -> Network:
        if self._overriden_active_network is not None:
            return self._overriden_active_network
        if boa.env.nickname in self._networks:
            return self._networks[boa.env.nickname]
        new_network = Network(
            name=boa.env.nickname, contracts=self._default_named_contracts
        )
        self._networks[new_network.name] = new_network
        return new_network

    def get_db_path(self) -> Path:
        return self.db_path

    def get_network(self, network_name_or_id: str | int) -> Network:
        if isinstance(network_name_or_id, int):
            return self.get_network_by_chain_id(network_name_or_id)
        else:
            if network_name_or_id.isdigit():
                return self.get_network_by_chain_id(int(network_name_or_id))
        return self.get_network_by_name(network_name_or_id)

    def get_network_by_chain_id(self, chain_id: int) -> Network:
        for network in self._networks.values():
            if network.chain_id == chain_id:
                return network
        raise ValueError(f"Network with chain_id {chain_id} not found.")

    def get_network_by_name(self, alias: str) -> Network:
        network = self._networks.get(alias, None)
        if not network:
            raise ValueError(f"Network {alias} not found.")
        return network

    def get_or_deploy_contract(
        self, *args, **kwargs
    ) -> VyperContract | ZksyncContract | ABIContract:
        return self.get_active_network().get_or_deploy_contract(*args, **kwargs)

    def set_active_network(
        self,
        name_of_network_or_network: str | Network,
        activate_boa=True,
        activate_db=True,
        **kwargs,
    ) -> Network:
        if not isinstance(name_of_network_or_network, str) and not isinstance(
            name_of_network_or_network, Network
        ):
            raise ValueError(
                f"The first argument must be a string or a Network object, you gave {name_of_network_or_network}."
            )

        if isinstance(name_of_network_or_network, str):
            name_of_network_or_network = self.get_network(name_of_network_or_network)
            name_of_network_or_network = cast(Network, name_of_network_or_network)

        name_of_network_or_network.set_kwargs(**kwargs)
        if activate_boa:
            name_of_network_or_network.create_and_set_or_set_boa_env()
            self._overriden_active_network = None
        else:
            self._overriden_active_network = name_of_network_or_network
        if activate_db:
            name_of_network_or_network._set_boa_db()
        self._networks[name_of_network_or_network.name] = name_of_network_or_network
        return name_of_network_or_network

    def activate_boa(self):
        active_network = self.get_active_network()
        active_network.create_and_set_or_set_boa_env()
        self._overriden_active_network = None

    def activate_db(self):
        active_network = self.get_active_network()
        active_network._set_boa_db()

    def _create_custom_network(self, url: str, is_fork: bool | None = False) -> Network:
        if is_fork is None:
            is_fork = False
        new_network = Network(
            name=f"custom_{self.custom_networks_counter}", url=url, is_fork=is_fork
        )
        self._networks[new_network.name] = new_network
        self.custom_networks_counter += 1
        return new_network

    @staticmethod
    def _validate_network_contracts_dict(
        contracts: Any, network_name: str | None = None
    ):
        network_name = f"{network_name}." if network_name else ""

        if not isinstance(contracts, dict):
            logger.error(f"networks.{network_name}contracts")
            raise ValueError(f"Contracts must be a dictionary in your {CONFIG_NAME}.")

        for contract in contracts:
            if not isinstance(contracts[contract], dict):
                logger.error(f"networks.contracts.{contract}")
                raise ValueError(
                    f"Contracts must be a dictionary in your {CONFIG_NAME}."
                )

    @staticmethod
    def _add_local_network_defaults(toml_data: dict) -> dict:
        if "networks" not in toml_data:
            toml_data["networks"] = {}

        for local_network, local_network_data in LOCAL_NETWORK_DEFAULTS.items():
            if local_network not in toml_data["networks"]:
                toml_data["networks"][local_network] = {}

            for key, value in local_network_data.items():
                if toml_data["networks"][local_network].get(key, None) is None:
                    toml_data["networks"][local_network][key] = value
        return toml_data

    @staticmethod
    def _add_fork_network_defaults(network_data: dict) -> dict:
        network_data["prompt_live"] = network_data.get(
            "prompt_live", FORK_NETWORK_DEFAULTS["prompt_live"]
        )
        network_data[SAVE_TO_DB] = network_data.get(
            SAVE_TO_DB, FORK_NETWORK_DEFAULTS[SAVE_TO_DB]
        )
        network_data["live_or_staging"] = network_data.get(
            "live_or_staging", FORK_NETWORK_DEFAULTS["live_or_staging"]
        )
        return network_data

    @staticmethod
    def _validate_fork_network_defaults(network_data: dict):
        if network_data.get(SAVE_TO_DB, False) is True:
            raise ValueError(
                "You cannot save forked network data to a live database. Please set 'save_to_db' to 'False' or leave it unset."
            )

    @staticmethod
    def _validate_local_network_data(network_data: dict, network_name: str):
        for key in network_data.keys():
            if key in RESTRICTED_VALUES_FOR_LOCAL_NETWORK:
                raise ValueError(f"Cannot set {key} for network {network_name}.")
            if network_name == PYEVM:
                if network_data.get("is_zksync", False) is True:
                    raise ValueError(
                        f"is_zksync for {network_name} must be false. Please adjust it in your config."
                    )
            if network_name == ERAVM:
                if network_data.get("is_zksync", True) is False:
                    raise ValueError(
                        f"is_zksync for {network_name} must be True. Please adjust it in your config."
                    )
            if network_data.get(SAVE_TO_DB) is True:
                raise ValueError(
                    f"{SAVE_TO_DB} for {network_name} must be 'False' or left unset. Please adjust it in your config."
                )


class Config:
    """A wrapper around the moccasin.toml file, it is also the main entry point for doing
    almost anything with moccasin.

    This class reads the moccasin.toml file and sets up project configuration.

    Attributes:
        _project_root (Path): The root directory of the project.
        _toml_data (dict): The raw data from moccasin.toml.
        dependencies (list[str]): Project dependencies from moccasin.toml.
        project (dict[str, str]): Project data from moccasin.toml.
        extra_data (dict[str, Any]): Any additional data from moccasin.toml.
        networks (_Networks): Network configurations.

    Args:
        root_path (Path, optional): The root directory of the project. Defaults to None.
    """

    _project_root: Path
    _toml_data: dict
    dependencies: list[str]
    project: dict[str, str]
    extra_data: dict[str, Any]
    networks: _Networks

    def __init__(self, root_path: Path | None):
        if root_path is None:
            root_path = Config.find_project_root()
        root_path = cast(Path, root_path)
        self._project_root = root_path
        self._toml_data = {}
        self.project = {}

        config_path: Path = root_path.joinpath(CONFIG_NAME)
        pyproject_path: Path = root_path.joinpath("pyproject.toml")

        if config_path.exists() or pyproject_path.exists():
            self._load_config(config_path, pyproject_path=pyproject_path)
        else:
            self._load_env_file()
            logger.warning(
                f"No {CONFIG_NAME} or pyproject.toml file found. Using default configuration."
            )
        self._initialize_networks()

    def _initialize_networks(self):
        db_path_str = self._toml_data.get("project", {}).get(
            "db_path", DB_PATH_LIVE_DEFAULT
        )
        db_path = Path(db_path_str).expanduser()
        if not db_path.is_absolute():
            db_path = self._project_root.joinpath(db_path)
        self.networks = _Networks(self._toml_data, db_path)

    def _load_config(self, config_path: Path, pyproject_path: Path | None = None):
        toml_data = self.read_configs(config_path, pyproject_path)
        # Need to get the .env file before expanding env vars
        self.project[DOT_ENV_KEY] = toml_data.get("project", {}).get(
            DOT_ENV_KEY, DOT_ENV_FILE
        )
        self._load_env_file()
        toml_data = self.expand_env_vars(toml_data)
        self.dependencies = toml_data.get("project", {}).get("dependencies", [])
        self.project = toml_data.get("project", {})
        self.extra_data = toml_data.get("extra_data", {})
        self._toml_data = toml_data
        if TESTS_FOLDER in self.project:
            logger.warning(
                f"Tests folder is set to {self.project[TESTS_FOLDER]}. This is not supported and will be ignored."
            )

    def _load_env_file(self):
        load_dotenv(dotenv_path=self.project_root.joinpath(self.dot_env))

    def get_config_path(self) -> Path:
        return self.config_path

    def read_configs_preserve_comments(
        self,
        moccasin_config_path: Path | None = None,
        pyproject_config_path: Path | None = None,
    ) -> tomlkit.TOMLDocument:
        if moccasin_config_path is None:
            moccasin_config_path = self.config_path
        if pyproject_config_path is None:
            pyproject_config_path = self.project_root.joinpath("pyproject.toml")
        moccasin_config = self.read_moccasin_toml_preserve_comments(
            moccasin_config_path
        )
        pyproject_config = self.read_pyproject_toml_preserve_comments(
            pyproject_config_path
        )
        if moccasin_config == {}:
            return pyproject_config
        merged_config = self.merge_configs(moccasin_config, pyproject_config)
        merged_config = cast(tomlkit.TOMLDocument, merged_config)
        return merged_config

    def read_configs(
        self, moccasin_path: Path | None = None, pyproject_path: Path | None = None
    ) -> dict:
        moccasin_data = self.read_moccasin_config(moccasin_path)
        pyproject_data = self.read_pyproject_config(pyproject_path)
        return self.merge_configs(moccasin_data, pyproject_data)

    def read_moccasin_config(self, config_path: Path | None = None) -> dict:
        if config_path is None:
            config_path = self.config_path
        return self.read_moccasin_toml(config_path)

    def read_pyproject_config(self, pyproject_path: Path | None = None) -> dict:
        if pyproject_path is None:
            pyproject_path = self.project_root.joinpath("pyproject.toml")
        return self.read_pyproject_toml(pyproject_path)

    def expand_env_vars(self, value):
        if isinstance(value, str):
            return os.path.expandvars(value)
        elif isinstance(value, dict):
            return {k: self.expand_env_vars(v) for k, v in value.items()}
        elif isinstance(value, list):
            return [self.expand_env_vars(item) for item in value]
        return value

    def get_networks(self) -> dict[str, Network]:
        return self.networks.get_networks()

    def get_active_network(self) -> Network:
        return self.networks.get_active_network()

    def get_db_path(self) -> Path:
        return self.networks.get_db_path()

    def get_or_deploy_contract(
        self, *args, **kwargs
    ) -> VyperContract | ZksyncContract | ABIContract:
        return self.get_active_network().get_or_deploy_contract(*args, **kwargs)

    def get_dependencies(self) -> list[str]:
        return self.dependencies

    def write_dependencies(self, dependencies: list):
        """Writes the dependencies to the config file.

        If a moccasin.toml file exists, it will write there, otherwise, it'll write to pyproject.toml.

        This will overwrite the existing dependencies with the new ones. So if you wish to keep old ones,
        read from the dependencies first.
        """
        toml_data = self.read_configs_preserve_comments()
        path_to_write = self.config_path
        if not self.config_path.exists() and self.pyproject_path.exists():
            self.nested_tomlkit_update(
                toml_data, ["tool", "moccasin", "project", "dependencies"], dependencies
            )
            path_to_write = self.pyproject_path
        else:
            self.nested_tomlkit_update(
                toml_data, ["project", "dependencies"], dependencies
            )

        # Create a temporary file in the same directory as the target file
        temp_file = tempfile.NamedTemporaryFile(
            mode="w",
            delete=False,
            dir=path_to_write.parent,
            prefix=".tmp_",
            suffix=".toml",
        )
        try:
            temp_file.write(tomlkit.dumps(toml_data))
            temp_file.close()
            shutil.move(temp_file.name, path_to_write)
        except Exception as e:
            os.unlink(temp_file.name)
            raise e
        self.dependencies = dependencies

    def get_base_dependencies_install_path(self) -> Path:
        project_root = self._project_root
        base_install_path = project_root / self.project.get(
            DEPENDENCIES_FOLDER, DEPENDENCIES_FOLDER
        )
        base_install_path.mkdir(exist_ok=True, parents=True)
        return base_install_path

    def get_root(self) -> Path:
        return self._project_root

    def find_contract(self, contract_or_contract_path: str) -> Path:
        return self._find_contract(
            self.project_root, self.contracts_folder, contract_or_contract_path
        )

    def set_active_network(
        self,
        name_url_or_id: str | Network,
        activate_boa=True,
        activate_db=True,
        **kwargs,
    ) -> Network:
        return self.networks.set_active_network(
            name_url_or_id, activate_boa=activate_boa, activate_db=activate_db, **kwargs
        )

    def activate_boa(self):
        self.networks.activate_boa()

    def activate_db(self):
        self.networks.activate_db()

    @property
    def config_path(self) -> Path:
        return self.project_root.joinpath(CONFIG_NAME)

    @property
    def pyproject_path(self) -> Path:
        return self.project_root.joinpath("pyproject.toml")

    @property
    def project_root(self) -> Path:
        return self._project_root

    @property
    def build_folder(self) -> str:
        return self.project.get(BUILD_FOLDER, BUILD_FOLDER)

    @property
    def out_folder(self) -> str:
        return self.build_folder

    @property
    def contracts_folder(self) -> str:
        return self.project.get(CONTRACTS_FOLDER, CONTRACTS_FOLDER)

    @property
    def src_folder(self) -> str:
        return self.contracts_folder

    @property
    def cov_config(self) -> str | None:
        return self.project.get("cov_config", None)

    @property
    def dot_env(self) -> str:
        return self.project.get(DOT_ENV_KEY, DOT_ENV_FILE)

    # Tests must be in "tests" folder
    @property
    def test_folder(self) -> str:
        return TESTS_FOLDER

    @property
    def script_folder(self) -> str:
        return self.project.get(SCRIPT_FOLDER, SCRIPT_FOLDER)

    @property
    def lib_folder(self) -> str:
        return self.project.get(DEPENDENCIES_FOLDER, DEPENDENCIES_FOLDER)

    @property
    def default_network(self) -> str:
        return self.networks.default_network_name

    @property
    def default_network_name(self) -> str:
        return self.default_network

    @staticmethod
    def load_config_from_root(project_root: Path | None = None) -> "Config":
        if project_root is None:
            project_root = Config.find_project_root()
        return Config(project_root)

    @staticmethod
    def find_project_root(start_path: Path | str | None = None) -> Path:
        if start_path is None:
            start_path = Path.cwd()
        start_path = Path(start_path).expanduser().resolve()
        current_path = start_path

        # Look for moccasin.toml
        while True:
            # Move up to the parent directory
            parent_path = current_path.parent
            if parent_path == current_path:
                # We've reached the root directory without finding moccasin.toml
                # raise FileNotFoundError(
                #     "Could not find moccasin.toml or src directory with Vyper contracts in any parent directory"
                # )
                break
            if (current_path / CONFIG_NAME).exists():
                return current_path
            # Check for src directory with .vy files in current directory
            src_path = current_path / "src"
            if src_path.is_dir() and any(src_path.glob("*.vy")):
                return current_path
            current_path = parent_path

        # Start over and look for pyproject.toml
        current_path = start_path
        while True:
            # Move up to the parent directory
            parent_path = current_path.parent
            if parent_path == current_path:
                # We've reached the root directory without finding moccasin.toml
                raise FileNotFoundError(
                    "Could not find moccasin.toml, pyproject.toml, or src directory with Vyper contracts in any parent directory"
                )
            if (current_path / "pyproject.toml").exists():
                return current_path
            current_path = parent_path

    @staticmethod
    def read_moccasin_toml(config_path: Path) -> dict:
        if not config_path.exists():
            return {}
        with open(config_path, "rb") as f:
            return tomllib.load(f)

    @staticmethod
    def read_moccasin_toml_preserve_comments(config_path: Path) -> tomlkit.TOMLDocument:
        config_path = Config._validated_moccasin_config_path(config_path)
        if not config_path.exists():
            return tomlkit.TOMLDocument()
        with open(config_path, "rb") as f:
            return tomlkit.load(f)

    @staticmethod
    def read_pyproject_toml_preserve_comments(
        config_path: Path,
    ) -> tomlkit.TOMLDocument:
        config_path = Config._validated_pyproject_config_path(config_path)
        if not config_path.exists():
            return tomlkit.TOMLDocument()
        with open(config_path, "rb") as f:
            return tomlkit.load(f)

    @staticmethod
    def read_pyproject_toml(pyproject_path: Path) -> dict:
        if not pyproject_path.exists():
            return {}

        with open(pyproject_path, "rb") as f:
            pyproject_data = tomllib.load(f)

        pyproject_config_toml = pyproject_data.get("tool", {}).get("moccasin", {})
        return pyproject_config_toml

    @staticmethod
    def merge_configs(
        moccasin_config_dict: Union[dict, tomlkit.TOMLDocument],
        pyproject_config_dict: Union[dict, tomlkit.TOMLDocument],
    ) -> Union[dict, tomlkit.TOMLDocument]:
        merged = moccasin_config_dict.copy()

        if (
            pyproject_config_dict.get("tool", {})
            .get("moccasin", {})
            .get("project", {})
            .get("dependencies", None)
            is not None
        ):
            logger.warning(
                "Dependencies in pyproject.toml will be overwritten by dependencies in moccasin.toml if the moccasin.toml has dependencies!"
            )

        def deep_update(d, u):
            for k, v in u.items():
                if isinstance(v, dict):
                    if k not in d:
                        d[k] = {}
                    deep_update(d[k], v)
                elif k not in d:
                    d[k] = v
            return d

        return deep_update(merged, pyproject_config_dict)

    @staticmethod
    def _validated_pyproject_config_path(config_path: Path):
        if not str(config_path).endswith("pyproject.toml"):
            config_path = config_path.joinpath("pyproject.toml")
        return config_path

    @staticmethod
    def _validated_moccasin_config_path(config_path: Path):
        if not str(config_path).endswith("moccasin.toml"):
            config_path = config_path.joinpath("moccasin.toml")
        return config_path

    @staticmethod
    def nested_tomlkit_update(
        toml_data: tomlkit.TOMLDocument,
        keys: list,
        value: str | int | float | bool | list | dict,
    ) -> tomlkit.TOMLDocument:
        current: Union[tomlkit.TOMLDocument, Table] = toml_data
        for key in keys[:-1]:
            if key not in current:
                current[key] = tomlkit.table()
            current = current[key]  # type: ignore
        current[keys[-1]] = value
        return toml_data

    @staticmethod
    def _find_contract(
        project_root: str | Path, contracts_folder: str, contract_or_contract_path: str
    ) -> Path:
        project_root = Path(project_root)
        # If the path starts with '~', expand to the user's home directory
        contract_path = Path(contract_or_contract_path).expanduser()

        # If the path is already absolute and exists, return it directly
        if contract_path.is_absolute() and contract_path.exists():
            return contract_path

        # Handle contract names without ".vy" by appending ".vy"
        if not contract_path.suffix == ".vy":
            contract_path = contract_path.with_suffix(".vy")

        # If the contract path is relative, check if it exists relative to config_root
        if not contract_path.is_absolute():
            contract_path = project_root / contract_path
            if contract_path.exists():
                return contract_path

        # Search for the contract in the contracts folder if not found by now
        contracts_location = project_root / contracts_folder
        contract_paths = list(contracts_location.rglob(contract_path.name))

        if not contract_paths:
            raise FileNotFoundError(
                f"Contract file '{contract_path.name}' not found under '{contracts_location}'."
            )
        elif len(contract_paths) > 1:
            found_paths = "\n".join(str(path) for path in contract_paths)
            raise FileExistsError(
                f"Multiple contract files named '{contract_path.name}' found:\n{found_paths}\n"
                "Please specify the full path to the contract file."
            )

        # Return the single found contract
        return contract_paths[0]


_config: Config | None = None


def get_active_network() -> Network:
    return get_config().get_active_network()


# REVIEW: Do we need... all of these?
def get_or_initialize_config(config_path: Path | None = None) -> Config:
    global _config
    if _config is None:
        _config = initialize_global_config(config_path)
    return _config


def get_config() -> Config:
    """Get the global Config object."""
    global _config
    if _config is None:
        raise ValueError(
            "Global Config object not initialized, initialize with initialize_global_config"
        )
    return _config


def initialize_global_config(config_path: Path | None = None) -> Config:
    global _config
    assert _config is None
    _set_global_config(config_path)
    return get_config()


def _set_global_config(config_path: Path | None = None) -> Config:
    global _config
    _config = Config.load_config_from_root(config_path)
    return _config



from moccasin._dependency_utils import (
    DependencyType,
    classify_dependency,
    preprocess_requirement,
)


def test_classify_dependency_pip():
    dep = "snekmate"
    assert classify_dependency(dep) == DependencyType.PIP


def test_classify_dependency_pip_version():
    dep = "snekmate==0.1.0"
    assert classify_dependency(dep) == DependencyType.PIP


def test_classify_dependency_pip_long():
    dep = '"git+https://github.com/pcaversaccio/snekmate.git"'
    assert classify_dependency(dep) == DependencyType.PIP


def test_classify_dependency_git():
    dep = "pcaversaccio/snekmate"
    assert classify_dependency(dep) == DependencyType.GITHUB


def test_classify_dependency_git_version():
    dep = "pcaversaccio/snekmate@0.1.0"
    assert classify_dependency(dep) == DependencyType.GITHUB


def test_classify_dependency_git_no_git():
    dep = "https://github.com/pcaversaccio/snekmate"
    assert classify_dependency(dep) == DependencyType.GITHUB


def test_preprocess_requirement():
    req = '"git+https://github.com/pcaversaccio/snekmate.git"'
    assert preprocess_requirement(req) == "snekmate"



import pytest

from moccasin.config import get_config


def pytest_configure(config):
    config.addinivalue_line(
        "markers",
        "staging: mark test to run on a live or staging environment, and all other tests to be skipped",
    )
    config.addinivalue_line(
        "markers",
        "local: mark test to run in a local or forked environment, all tests are implicitly marked as this",
    )
    config.addinivalue_line("markers", "local: all tests are implicitly marked as ")


def pytest_collection_modifyitems(config, items):
    moccasin_config = get_config()
    active_network = moccasin_config.get_active_network()
    if active_network.live_or_staging:
        skip_non_staging = pytest.mark.skip(reason="Not a staging test")
        for item in items:
            if "staging" not in item.keywords:
                item.add_marker(skip_non_staging)
    else:
        skip_staging = pytest.mark.skip(reason="Not running on a live network")
        for item in items:
            if "staging" in item.keywords and "local" not in item.keywords:
                item.add_marker(skip_staging)



import os
import subprocess
from pathlib import Path

from tests.conftest import COMPLEX_PROJECT_PATH


def test_run_help(mox_path):
    current_dir = Path.cwd()
    try:
        os.chdir(COMPLEX_PROJECT_PATH)
        result = subprocess.run(
            [mox_path, "wallet", "-h"], check=True, capture_output=True, text=True
        )
    finally:
        os.chdir(current_dir)
    assert "Moccasin CLI wallet" in result.stdout



import os

import boa
from boa.verifiers import Blockscout
from contracts import Counter


def moccasin_main():
    counter = Counter.at("0x64A5F381A9D4eBA6d6A7D24De4D799901B19d29d")
    api_key = os.getenv("BLOCKSCOUT_API_KEY")
    blockscout = Blockscout("https://eth-sepolia.blockscout.com", api_key)
    with boa.set_verifier(blockscout):
        result = boa.verify(counter)
        result.wait_for_verification()
        assert result.is_verified()



from moccasin.config import get_active_network


def moccasin_main():
    active_network = get_active_network()
    contract = active_network.get_latest_contract_checked("Difficulty")
    print("Contract is: ", contract)



import contextlib
import sys
from pathlib import Path
from typing import Iterator, List

import boa
from boa.util.abi import Address

from moccasin.config import Config, Network, get_config
from moccasin.constants.vars import (
    ERA_DEFAULT_PRIVATE_KEY,
    ERAVM,
    GITHUB,
    PYPI,
    STARTING_BOA_BALANCE,
)
from moccasin.logging import logger
from moccasin.moccasin_account import MoccasinAccount


def get_sys_paths_list(config: Config) -> List[Path]:
    config_root = config.get_root()
    config_contracts = config_root.joinpath(config.contracts_folder)

    base_install_path = config_root.joinpath(config.lib_folder)

    # REVIEW: We could also look into the versions.toml file and add those to the search path too
    # This way, we could make the imports even cleaner
    github_dependencies = base_install_path.joinpath(GITHUB)
    pypi_dependencies = base_install_path.joinpath(PYPI)

    return [
        config_root,
        config_contracts,
        github_dependencies,
        pypi_dependencies,
        base_install_path,
    ]


@contextlib.contextmanager
def _patch_sys_path(paths: List[Path]) -> Iterator[None]:
    str_paths = [str(p) for p in paths]
    anchor = sys.path
    try:
        # add these with highest precedence -- conflicts should prefer user modules/code
        sys.path = str_paths + sys.path
        yield
    finally:
        sys.path = anchor


# REVIEW: Might be best to just set this as **kwargs
def _get_set_active_network_from_cli_and_config(
    config: Config,
    network: str = None,
    url: str = None,
    fork: bool | None = None,
    account: str | None = None,
    password_file_path: Path | None = None,
    prompt_live: bool | None = None,
    explorer_uri: str | None = None,
    explorer_api_key: str | None = None,
    db_path: str | None = None,
    save_to_db: bool | None = None,
) -> Network:
    if network is None:
        network = config.default_network

    config.set_active_network(
        network,
        is_fork=fork,
        url=url,
        default_account_name=account,
        # private_key=private_key, # No private key in networks
        # password=password, # No password in networks
        password_file_path=password_file_path,
        prompt_live=prompt_live,
        explorer_uri=explorer_uri,
        explorer_api_key=explorer_api_key,
        db_path=db_path,
        save_to_db=save_to_db,
    )

    active_network: Network = config.get_active_network()
    logger.debug(f"Active network set to: {active_network.name}")
    if active_network is None:
        raise ValueError("No active network set. Please set a valid network.")
    return active_network


def _setup_network_and_account_from_config_and_cli(
    network: str = None,
    url: str = None,
    fork: bool | None = None,
    account: str | None = None,
    private_key: str | None = None,
    password: str | None = None,
    password_file_path: Path | None = None,
    prompt_live: bool | None = None,
    explorer_uri: str | None = None,
    explorer_api_key: str | None = None,
    db_path: str | None = None,
    save_to_db: bool | None = None,
):
    """All the network and account logic in the function parameters are from the CLI.
    We will use the order of operations to setup the network:

        1. scripts (which, we don't touch here)
        2. CLI
        3. Config
        4. Default Values

    All the values passed into this function come from the CLI.
    """
    if account is not None and private_key is not None:
        raise ValueError("Cannot set both account and private key in the CLI!")

    mox_account: MoccasinAccount | None = None
    config = get_config()

    # 1. Update the network with the CLI values
    active_network = _get_set_active_network_from_cli_and_config(
        config,
        network,
        url,
        fork,
        account,
        password_file_path,
        prompt_live,
        explorer_uri,
        explorer_api_key,
        db_path,
        save_to_db,
    )

    # 2. Update and set account
    if active_network.prompt_live:
        if not fork:
            response = input(
                "The transactions run on this will actually be broadcast/transmitted, spending gas associated with your account. Are you sure you wish to continue?\nType 'y' or 'Y' and hit 'ENTER' or 'RETURN' to continue:\n"
            )
            if response.lower() != "y":
                logger.info("Operation cancelled.")
                sys.exit(0)

    if active_network.default_account_name and private_key is None:
        # This will also attempt to unlock the account with a prompt
        # If no password or password file is passed
        mox_account = MoccasinAccount(
            keystore_path_or_account_name=active_network.default_account_name,
            password=password,
            password_file_path=active_network.unsafe_password_file,
        )

    # Private key overrides the default account
    if private_key:
        mox_account = MoccasinAccount(
            private_key=private_key,
            password=password,
            password_file_path=active_network.unsafe_password_file,
        )

    if mox_account:
        if active_network.is_local_or_forked_network():
            boa.env.eoa = mox_account.address
        else:
            boa.env.add_account(mox_account, force_eoa=True)

    if not mox_account and active_network.name is ERAVM:
        boa.env.add_account(MoccasinAccount(private_key=ERA_DEFAULT_PRIVATE_KEY))

    # Check if it's a fork, pyevm, or eravm
    if not active_network.is_local_or_forked_network():
        if boa.env.eoa is None:
            logger.warning(
                "No default EOA account found. Please add an account to the environment before attempting a transaction."
            )

    if isinstance(boa.env.eoa, Address) and active_network.is_local_or_forked_network():
        boa.env.set_balance(boa.env.eoa, STARTING_BOA_BALANCE)



import logging


class CustomFormatter(logging.Formatter):
    def format(self, record):
        if record.levelno == logging.WARNING:
            return f"WARNING: {record.getMessage()}"
        elif record.levelno == logging.ERROR:
            return f"ERROR: {record.getMessage()}"
        return record.getMessage()


logger = logging.getLogger("titanoboa")
ch = logging.StreamHandler()
ch.setFormatter(CustomFormatter())
logger.addHandler(ch)


def set_log_level(quiet=False, debug=False):
    if debug:
        logger.setLevel(logging.DEBUG)
    elif not quiet:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.ERROR)



import atexit
import code
import readline
from argparse import Namespace
from pathlib import Path

from moccasin._sys_path_and_config_setup import (
    _patch_sys_path,
    _setup_network_and_account_from_config_and_cli,
    get_sys_paths_list,
)
from moccasin.config import initialize_global_config
from moccasin.constants.vars import CONSOLE_HISTORY_FILE, MOCCASIN_DEFAULT_FOLDER
from moccasin.logging import logger


def main(args: Namespace) -> int:
    config = initialize_global_config()

    # Set up the environment (add necessary paths to sys.path, etc.)
    with _patch_sys_path(get_sys_paths_list(config)):
        _setup_network_and_account_from_config_and_cli(
            network=args.network,
            url=args.url,
            fork=args.fork,
            account=args.account,
            private_key=args.private_key,
            password=args.password,
            password_file_path=args.password_file_path,
            prompt_live=args.prompt_live,
            db_path=args.db_path,
        )

        # Ensure the Moccasin folder exists
        MOCCASIN_DEFAULT_FOLDER.mkdir(parents=True, exist_ok=True)

        history_file = MOCCASIN_DEFAULT_FOLDER.joinpath(CONSOLE_HISTORY_FILE)
        validate_history_file(history_file)
        history_file_str = str(history_file)

        # Enable history
        readline.parse_and_bind("tab: complete")

        # Try to read the history file
        try:
            readline.read_history_file(history_file_str)
        except FileNotFoundError:
            print("No history file found. Starting with empty history.")
        except Exception as e:
            print(f"Error reading history file: {e}")
            print("Starting with empty history.")

        # Set history length
        readline.set_history_length(1000)

        # Save history on exit
        atexit.register(readline.write_history_file, history_file_str)

        local_vars = globals().copy()
        local_vars.update(locals())
        console = MoccasinConsole(local_vars)

        # Run the console
        console.interact(
            "Welcome to the interactive console! Your command history will be saved."
        )
    return 0


def validate_history_file(history_file: Path) -> None:
    HISTORY_HEADER = "_HiStOrY_V2_\n"
    MIN_VALID_SIZE = 10

    if not history_file.exists():
        logger.debug("History file does not exist. A new one will be created.")
        return

    file_size = history_file.stat().st_size
    logger.debug(f"History file exists. Size: {file_size} bytes")

    if file_size < MIN_VALID_SIZE:
        logger.debug("History file is too small. Appending header.")
        with history_file.open("a") as f:
            f.write(HISTORY_HEADER)


class MoccasinConsole(code.InteractiveConsole):
    def __init__(self, locals=None, filename="<console>"):
        super().__init__(locals, filename)
        self.exit_requested = False

    def raw_input(self, prompt=""):
        line = input(prompt)
        if line.strip().lower() == "q":
            self.exit_requested = True
            raise EOFError("Quit requested")
        return line

    def interact(self, banner=None, exitmsg=None):
        try:
            super().interact(banner, exitmsg)
        except EOFError:
            if self.exit_requested:
                print("Exiting console...")
            else:
                raise



def test_increment_one(counter_contract):
    counter_contract.increment()
    assert counter_contract.number() == 2


def test_increment_two(counter_contract):
    counter_contract.increment()
    counter_contract.increment()
    assert counter_contract.number() == 3



from contracts import Counter

from moccasin.boa_tools import VyperContract


def deploy() -> VyperContract:
    counter: VyperContract = Counter.deploy()
    print("Starting count: ", counter.number())
    counter.increment()
    print("Ending count: ", counter.number())
    return counter


def moccasin_main() -> VyperContract:
    return deploy()



import pytest
from script.deploy import deploy
from script.deploy_coffee import deploy as deploy_coffee
from script.mock_deployer.deploy_feed import deploy_mock


@pytest.fixture
def price_feed():
    return deploy_mock()


@pytest.fixture
def eth_usd():
    return deploy_mock()


@pytest.fixture
def counter_contract():
    return deploy()


@pytest.fixture
def coffee():
    return deploy_coffee()



import json
import os
import subprocess
from pathlib import Path

import pytest

from moccasin.commands.compile import compile_
from moccasin.commands.run import run_script
from tests.conftest import ZKSYNC_PROJECT_PATH


def test_compile_zksync_pyevm(zksync_cleanup_out_folder, zksync_out_folder, mox_path):
    current_dir = Path.cwd()
    try:
        os.chdir(current_dir.joinpath(ZKSYNC_PROJECT_PATH))
        result = subprocess.run(
            [mox_path, "build", "Difficulty.vy", "--network", "pyevm"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert "Done compiling Difficulty" in result.stderr
    assert result.returncode == 0
    # read the Difficulty.json in zksync_out_folder
    with open(
        ZKSYNC_PROJECT_PATH.joinpath(zksync_out_folder).joinpath("Difficulty.json"), "r"
    ) as f:
        data = json.load(f)
        assert data["vm"] == "evm"


def test_compile_zksync_one(
    zksync_cleanup_out_folder, zksync_out_folder, zksync_test_env
):
    compile_(
        ZKSYNC_PROJECT_PATH.joinpath("src/Difficulty.vy"),
        ZKSYNC_PROJECT_PATH.joinpath(zksync_out_folder),
        is_zksync=True,
        write_data=True,
    )
    with open(
        ZKSYNC_PROJECT_PATH.joinpath(zksync_out_folder).joinpath("Difficulty.json"), "r"
    ) as f:
        data = json.load(f)
        assert data["vm"] == "eravm"


def test_compile_zksync_bad(
    zksync_cleanup_out_folder, zksync_out_folder, zksync_test_env
):
    with pytest.raises(AssertionError) as excinfo:
        compile_(
            ZKSYNC_PROJECT_PATH.joinpath("src/SelfDestruct.vy"),
            ZKSYNC_PROJECT_PATH.joinpath(zksync_out_folder),
            is_zksync=True,
            write_data=True,
        )
    error_message = str(excinfo.value)
    assert "subprocess compiling" in error_message
    assert "failed with exit code Some(1)" in error_message
    assert "The `SELFDESTRUCT` instruction is not supported" in error_message


def test_run_zksync_good(zksync_cleanup_out_folder, zksync_test_env):
    difficulty_contract = run_script(ZKSYNC_PROJECT_PATH.joinpath("script/deploy.py"))
    difficulty = difficulty_contract.get_difficulty()
    assert difficulty == 2500000000000000



import pprint
from argparse import Namespace
from typing import Any

from moccasin.commands.compile import compile_
from moccasin.config import Config, get_config, initialize_global_config

FUNCTION_SIGNATURES_ALTS = [
    "methods",
    "signatures",
    "selectors",
    "function-selectors",
    "function_signatures",
]


def main(args: Namespace) -> int:
    inspect_type = args.inspect_type.replace("-", "_")
    config = initialize_global_config()
    inspect_contract(args.contract, inspect_type, config=config, print_out=True)
    return 0


def inspect_contract(
    contract: str,
    inspect_type: str,
    config: Config | None = None,
    print_out: bool = False,
) -> Any:
    if config is None:
        config = get_config()

    contract_path = config.find_contract(contract)
    vyper_deployer = compile_(
        contract_path, config.get_root().joinpath(config.build_folder)
    )

    if inspect_type in FUNCTION_SIGNATURES_ALTS:
        inspect_type = "function_signatures"

    if vyper_deployer is None:
        raise FileNotFoundError(
            f"Could not compile contract '{contract_path}'. Please check the contract file."
        )
    inspected_data = getattr(vyper_deployer.compiler_data, inspect_type)
    final_data = inspected_data

    if inspect_type == "function_signatures":
        final_data = {}
        for _, contract_function in inspected_data.items():
            for signature in contract_function.method_ids:
                int_selector = contract_function.method_ids[signature]
                final_data[signature] = f"{hex(int_selector)} ({int_selector})"

    if print_out:
        print(f"Signatures and selectors for {contract_path.stem}:")
        if inspect_type == "function_signatures":
            for function, selector in final_data.items():
                pprint.pprint(f"{function}: {selector}")
        else:
            pprint.pprint(final_data)
    return final_data



def test_using_fixture_one(price_feed):
    assert price_feed.address is not None


def test_using_fixture_two(price_feed, eth_usd):
    assert price_feed.address is not None
    assert eth_usd.address is not None



from dataclasses import dataclass
from pathlib import Path
from typing import Any

import boa
from boa.contracts.vyper.vyper_contract import VyperContract, VyperDeployer
from boa.util.abi import Address
from boa_zksync.contract import ZksyncContract

from moccasin.logging import logger


@dataclass
class NamedContract:
    contract_name: str
    force_deploy: bool | None = None
    abi: str | None = None
    abi_from_explorer: bool | None = None
    deployer_script: str | Path | None = None
    address: str | None = None
    vyper_contract: VyperContract | None = None
    vyper_deployer: VyperDeployer | None = None

    def _reset(self):
        self.vyper_contract = None

    def update_from_deployed_contract(self, deployed_contract: VyperContract):
        self.abi = deployed_contract.abi
        self.address = deployed_contract.address
        self.vyper_contract = deployed_contract
        self.vyper_deployer = deployed_contract.deployer

    def set_defaults(self, other: "NamedContract"):
        self.force_deploy = (
            self.force_deploy if self.force_deploy is not None else other.force_deploy
        )
        self.abi = self.abi if self.abi is not None else other.abi
        self.abi_from_explorer = (
            self.abi_from_explorer
            if self.abi_from_explorer is not None
            else other.abi_from_explorer
        )
        self.deployer_script = (
            self.deployer_script
            if self.deployer_script is not None
            else other.deployer_script
        )
        self.address = self.address if self.address is not None else other.address

    def get(self, key: str, otherwise: Any):
        return getattr(self, key, otherwise)

    def is_active(self) -> bool:
        if self.address is not None:
            boa_contract = boa.env._contracts.get(
                Address(self.address).canonical_address, None
            )
            if boa_contract is None:
                return False
            if boa_contract == self.vyper_contract:
                return True
        return False

    def _deploy(
        self,
        script_folder: str,
        deployer_script: str | Path | None = None,
        update_from_deploy: bool = True,
    ) -> VyperContract | ZksyncContract:
        if deployer_script:
            deployer_script = str(deployer_script)
            deployer_module_path = (
                deployer_script
                if deployer_script.startswith(script_folder)
                else f"{script_folder}.{deployer_script}"
            )
        deployer_script = (
            self.deployer_script if deployer_script is None else deployer_script
        )

        if not deployer_script:
            raise ValueError("Deployer path not provided")

        deployer_module_path = deployer_module_path.replace("/", ".")
        deployer_module_path = (
            deployer_module_path[:-3]
            if deployer_module_path.strip().endswith(".py")
            else deployer_module_path
        )
        logger.debug(f"Deploying contract using {deployer_module_path}...")
        import importlib

        vyper_contract: VyperContract | ZksyncContract = importlib.import_module(
            f"{deployer_module_path}"
        ).moccasin_main()

        if not isinstance(vyper_contract, VyperContract) and not isinstance(
            vyper_contract, ZksyncContract
        ):
            raise ValueError(
                f"Your {deployer_module_path} script for {self.contract_name} set in deployer path must return a VyperContract or ZksyncContract object"
            )
        if update_from_deploy:
            self.update_from_deployed_contract(vyper_contract)
        return vyper_contract

    def _from_deploy_script(self):
        pass



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

import moccasin

sys.path.insert(0, os.path.abspath("."))

project = "moccasin"
copyright = "2024, Patrick Collins, Charles Cooper"
author = "Patrick Collins, Charles Cooper"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinxarg.ext",
    "sphinx_copybutton",
    "sphinx_multiversion",
    "sphinxarg.ext",
    "sphinx_tabs.tabs",
]

master_doc = "toctree"
templates_path = ["_templates"]
exclude_patterns = [".DS_Store", "_autogenerated/*.rst", "coming_soon.rst"]

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "shibuya"
html_theme_options = {
    "accent_color": "orange",
    "twitter_creator": "cyfrinaudits",
    "twitter_site": "cyfrinaudits",
    "twitter_url": "https://twitter.com/cyfrinaudits",
    "github_url": "https://github.com/cyfrin",
}
html_css_files = ["custom.css"]

html_static_path = ["_static"]
html_favicon = "_static/favicon.svg"

# For the "Edit this page ->" link
html_context = {
    "source_type": "github",
    "source_user": "cyfrin",
    "source_repo": "moccasin",
}

# Options for sphinx_multiversion
smv_remote_whitelist = r"^origin$"
smv_branch_whitelist = r"^master$"  # master is gross
smv_tag_whitelist = r"^v\d+\.\d+.\d+$"
version = moccasin.version()

jinja_contexts = {
    "moccasin_wrapper_for_docs": {
        "get_mox_help_output": "moccasin_wrapper_for_docs.get_mox_help_output"
    }
}



from contracts import BuyMeACoffee

from moccasin.boa_tools import VyperContract
from moccasin.config import get_config


def deploy() -> VyperContract:
    active_network = get_config().get_active_network()
    price_feed: VyperContract = active_network.get_or_deploy_contract("price_feed")
    buy_me_a_coffe: VyperContract = BuyMeACoffee.deploy(price_feed.address)
    print(f"Deployed BuyMeACoffee to {buy_me_a_coffe.address}")
    return buy_me_a_coffe


def moccasin_main() -> VyperContract:
    return deploy()



import os
import subprocess
import warnings
from pathlib import Path

from moccasin.commands.run import run_script

MOCK_AGGREGATOR = "MockV3Aggregator"
COUNTER = "Counter"


# ------------------------------------------------------------------
#                READING AND WRITING DEPLOYMENTS
# ------------------------------------------------------------------
def test_local_networks_dont_have_data_saved_to_db(
    deployments_config, deployments_path, anvil
):
    current_dir = Path.cwd()
    starting_deployments_number = 0
    try:
        os.chdir(deployments_path)
        run_script("deploy")
    finally:
        os.chdir(current_dir)
    assert starting_deployments_number == 0

    active_network = deployments_config.get_active_network()

    active_network.get_latest_contract_unchecked(COUNTER)
    active_network.get_deployments_unchecked(COUNTER)


def test_checks_integrity_of_contracts(
    mox_path, deployments_config, deployments_path, deployments_contract_override, anvil
):
    current_dir = Path.cwd()
    try:
        os.chdir(deployments_path)
        deployments_config.set_active_network("anvil", activate_boa=False)
        active_network = deployments_config.get_active_network()
        anvil_chain_id = 31337

        starting_deployments_number_checked = len(
            active_network.get_deployments_checked(
                contract_name=COUNTER, limit=None, chain_id=anvil_chain_id
            )
        )
        starting_deployments_number_unchecked = len(
            active_network.get_deployments_unchecked(
                contract_name=COUNTER, limit=None, chain_id=anvil_chain_id
            )
        )

        result = subprocess.run(
            [mox_path, "run", "deploy", "--network", "anvil"],
            check=True,
            text=True,
            capture_output=True,
        )
    finally:
        os.chdir(current_dir)

    assert "Ending count:  2" in result.stdout
    assert starting_deployments_number_unchecked == 2
    assert starting_deployments_number_checked == 0

    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", message="Requested .* but there is no bytecode at that address!"
        )
        latest_contract = active_network.get_latest_contract_checked(
            contract_name=COUNTER
        )
    checked_deployments = active_network.get_deployments_checked(contract_name=COUNTER)
    unchecked_deployments = active_network.get_deployments_unchecked(
        contract_name=COUNTER
    )

    # Ending deployments number should be 1, since they don't have the same integrity
    assert len(checked_deployments) == 1
    assert len(unchecked_deployments) == 3
    assert latest_contract.address is not None


def test_records_deployment_on_deployment(
    mox_path, deployments_config, deployments_path, anvil
):
    current_dir = Path.cwd()
    starting_deployments_number = 0
    try:
        os.chdir(deployments_path)
        deployments_config.set_active_network("anvil", activate_boa=False)
        active_network = deployments_config.get_active_network()
        starting_deployments_number = len(
            active_network.get_deployments_unchecked(contract_name=COUNTER)
        )
        result = subprocess.run(
            [mox_path, "run", "deploy", "--network", "anvil"],
            check=True,
            text=True,
            capture_output=True,
        )
    finally:
        os.chdir(current_dir)
    assert starting_deployments_number == 2

    assert "Ending count:  1" in result.stdout

    active_network = deployments_config.get_active_network()
    deployments = active_network.get_deployments_unchecked(COUNTER)

    assert len(deployments) == 3



from argparse import Namespace
from pathlib import Path

from moccasin.constants.file_data import (
    CONFTEST_DEFAULT,
    COUNTER_VYPER_CONTRACT_SRC,
    COVERAGERC,
    DEPLOY_SCRIPT_DEFAULT,
    GITATTRIBUTES,
    GITIGNORE,
    MOCCASIN_DEFAULT_CONFIG,
    PYPROJECT_TOML_DEFAULT,
    README_MD_SRC,
    TEST_COUNTER_DEFAULT,
    VSCODE_SETTINGS_DEFAULT,
)
from moccasin.constants.vars import (
    CONFIG_NAME,
    CONTRACTS_FOLDER,
    COUNTER_CONTRACT,
    DEFAULT_PROJECT_FOLDERS,
    README_PATH,
    SCRIPT_FOLDER,
    TESTS_FOLDER,
)
from moccasin.logging import logger


def main(args: Namespace) -> int:
    path: Path = new_project(
        args.path or ".",
        args.force or False,
        args.vscode or False,
        args.pyproject or False,
    )
    logger.info(f"Project initialized at {str(path)}")
    return 0


def new_project(
    project_path_str: str = ".",
    force: bool = False,
    vscode: bool = False,
    pyproject: bool = False,
) -> Path:
    """Initializes a new project.

    Args:
        project_path: Path to initialize the project at. If not exists, it will be created.
        force: If True, it will create folders and files even if the folder is not empty.
        vscode: If True, it will create a .vscode folder with settings.json

    Returns the path to the project as a string.
    """
    project_path = Path(project_path_str).resolve()
    if not force and project_path.exists() and list(project_path.glob("*")):
        raise FileExistsError(
            f"Directory is not empty: {project_path}.\nIf you're sure the folder is ok to potentially overwrite, try creating a new project by running with `mox init --force`"
        )
    project_path.mkdir(exist_ok=True)
    _create_folders(project_path, vscode=vscode)
    _create_files(project_path, vscode=vscode, pyproject=pyproject)
    return project_path


def _create_folders(project_path: Path, vscode: bool = False) -> None:
    for folder in DEFAULT_PROJECT_FOLDERS:
        Path(project_path).joinpath(folder).mkdir(exist_ok=True, parents=True)
    if vscode:
        Path(project_path).joinpath(".vscode").mkdir(exist_ok=True)


def _create_files(
    project_path: Path, vscode: bool = False, pyproject: bool = False
) -> None:
    _write_file(project_path.joinpath(".gitignore"), GITIGNORE)
    _write_file(project_path.joinpath(".gitattributes"), GITATTRIBUTES)
    _write_file(project_path.joinpath(".coveragerc"), COVERAGERC)
    _write_file(
        project_path.joinpath(f"{CONTRACTS_FOLDER}/{COUNTER_CONTRACT}"),
        COUNTER_VYPER_CONTRACT_SRC,
    )
    _write_file(project_path.joinpath(CONFIG_NAME), MOCCASIN_DEFAULT_CONFIG)
    _write_file(project_path.joinpath(README_PATH), README_MD_SRC)
    _write_file(project_path.joinpath(f"{TESTS_FOLDER}/conftest.py"), CONFTEST_DEFAULT)
    _write_file(
        project_path.joinpath(f"{TESTS_FOLDER}/test_counter.py"), TEST_COUNTER_DEFAULT
    )
    _write_file(
        project_path.joinpath(f"{SCRIPT_FOLDER}/deploy.py"), DEPLOY_SCRIPT_DEFAULT
    )
    _write_file(project_path.joinpath(f"{SCRIPT_FOLDER}/__init__.py"), "")
    if vscode:
        _write_file(
            project_path.joinpath(".vscode/settings.json"), VSCODE_SETTINGS_DEFAULT
        )
    if pyproject:
        _write_file(project_path.joinpath("pyproject.toml"), PYPROJECT_TOML_DEFAULT)


def _write_file(path: Path, contents: str, overwrite: bool = False) -> None:
    if not path.exists() or overwrite:
        with path.open("w") as fp:
            fp.write(contents)



import argparse
from typing import Tuple

from moccasin.__main__ import generate_main_parser_and_sub_parsers


# ------------------------------------------------------------------
#                          CLI COMMANDS
# ------------------------------------------------------------------
# Wrapper function for sphinx-argparse
def get_main_parser() -> argparse.ArgumentParser:
    main_parser, _ = generate_main_parser_and_sub_parsers()
    return main_parser


def get_parsers() -> Tuple[argparse.ArgumentParser, dict[str, argparse.ArgumentParser]]:
    main_parser, subparsers_action = generate_main_parser_and_sub_parsers()
    subparsers = {
        choice: subparsers_action.choices[choice]  # type: ignore
        for choice in subparsers_action.choices  # type: ignore
    }
    return main_parser, subparsers


def get_subparser(name: str) -> argparse.ArgumentParser:
    _, subparsers = get_parsers()
    return subparsers[name]


def get_init():
    return get_subparser("init")


def get_compile():
    return get_subparser("compile")


def get_run():
    return get_subparser("run")


def get_deploy():
    return get_subparser("deploy")


def get_deployments():
    return get_subparser("deployments")


def get_install():
    return get_subparser("install")


def get_test():
    return get_subparser("test")


def get_wallet():
    return get_subparser("wallet")


def get_console():
    return get_subparser("console")


def get_purge():
    return get_subparser("purge")


def get_config():
    return get_subparser("config")


def get_explorer():
    return get_subparser("explorer")


def get_inspect():
    return get_subparser("inspect")



import re
from dataclasses import dataclass
from enum import Enum

from packaging.requirements import InvalidRequirement, Requirement

from moccasin.config import get_or_initialize_config
from moccasin.logging import logger


class DependencyType(Enum):
    GITHUB = "github"
    PIP = "pip"


def classify_dependency(dependency: str) -> DependencyType:
    dependency = dependency.strip().strip("'\"")

    github_shorthand = r"^([a-zA-Z0-9_-]+)/([a-zA-Z0-9_-]+)(@[a-zA-Z0-9_.-]+)?$"
    github_url = r"^https://github\.com/[a-zA-Z0-9_-]+/[a-zA-Z0-9_-]+$"

    if re.match(github_shorthand, dependency) or re.match(github_url, dependency):
        return DependencyType.GITHUB

    return DependencyType.PIP


def _write_new_dependencies(
    new_package_ids: list[str], dependency_type: DependencyType
):
    config = get_or_initialize_config()
    dependencies = config.get_dependencies()

    typed_dependencies = [
        preprocess_requirement(dep)
        for dep in dependencies
        if classify_dependency(dep) == dependency_type
    ]

    to_delete = set()
    updated_packages = set()

    if dependency_type == DependencyType.PIP:
        for package in new_package_ids:
            try:
                processed_package = preprocess_requirement(package)
                package_req = Requirement(processed_package)
            except InvalidRequirement:
                logger.warning(f"Invalid requirement format for package: {package}")
                continue
            for dep in typed_dependencies:
                dep_req = Requirement(dep)
                if dep_req.name == package_req.name:
                    to_delete.add(dep)
                    updated_packages.add(package_req.name)

            if package_req.name not in updated_packages:
                logger.info(f"Installed new package: {package}")
            else:
                logger.info(f"Updated package: {package}")
    else:  # GIT dependencies
        for package in new_package_ids:
            package_dep = GitHubDependency.from_string(package)
            for dep in typed_dependencies:
                dep_gh = GitHubDependency.from_string(dep)
                if dep_gh.org == package_dep.org and dep_gh.repo == package_dep.repo:
                    to_delete.add(dep)
                    updated_packages.add(package_dep.format_no_version())

            if f"{package_dep.org}/{package_dep.repo}" not in updated_packages:
                logger.info(f"Installed {package}")
            else:
                logger.info(f"Updated {package}")

    # Remove old versions of updated packages
    dependencies = [dep for dep in dependencies if dep not in to_delete]

    # Add new packages while preserving order
    new_deps = []
    for dep in dependencies + new_package_ids:
        if dep not in new_deps:
            new_deps.append(dep)

    if len(new_deps) > 0:
        config.write_dependencies(new_deps)


def preprocess_requirement(package: str) -> str:
    package = package.strip().strip("'\"")
    git_url_pattern = r"^git\+https?://.*"
    if re.match(git_url_pattern, package):
        package_name = package.split("/")[-1].replace(".git", "")
        return package_name
    return package


@dataclass
class GitHubDependency:
    org: str
    repo: str
    version: str | None = None

    @classmethod
    def from_string(cls, dep_string: str) -> "GitHubDependency":
        if "@" in dep_string:
            path, version = dep_string.split("@")
        else:
            path, version = dep_string, None

        org, repo = str(path).split("/")
        org = org.strip().lower()
        repo = repo.strip().lower()
        version = version.strip() if version else None
        return cls(org, repo, version)

    def format_no_version(self) -> str:
        return f"{self.org}/{self.repo}"

    def __str__(self) -> str:
        if self.version:
            return f"{self.org}/{self.repo}@{self.version}"
        return self.format_no_version()



from argparse import Namespace
from enum import Enum

from boa.deployments import Deployment

from moccasin.config import Config, get_config, initialize_global_config
from moccasin.logging import logger

NUM_DASH = 60


class PrintVerbosity(Enum):
    CONTRACT_ADDRESS = 0  # Only the contract address
    CONTRACT_ADDRESS_AND_NAME = 1  # Contract address and contract name
    ADDRESS_NAME_DEPLOYER = 2  # Address, name, and deployer address
    FULL_DETAILS = 3  # All available details
    RAW = 4  # Raw deployment object


def main(args: Namespace) -> int:
    initialize_global_config()
    print_deployments_from_cli(
        args.contract_name,
        args.format_level,
        args.db_path,
        args.checked,
        args.limit,
        args.network,
        args.url,
        args.fork,
    )
    return 0


def print_deployments_from_cli(
    contract_name: str | None = None,
    format_level: int = PrintVerbosity.CONTRACT_ADDRESS_AND_NAME.value,
    db_path: str | None = None,
    checked: bool = False,
    limit: int | None = None,
    network: str = None,
    url: str = None,
    fork: bool = None,
    config: Config | None = None,
) -> list[Deployment]:
    if config is None:
        config = get_config()

    if contract_name is not None:
        if contract_name.strip().lower() == "all":
            contract_name = None

    if network is None:
        network = config.default_network

    # Don't activate boa
    config.set_active_network(
        network, url=url, fork=fork, db_path=db_path, activate_boa=False
    )
    active_network = config.get_active_network()

    if not hasattr(active_network, "chain_id"):
        logger.warning(
            "No chain_id found in active network, attempting to connect to the network to retrieve the Chain Id"
        )
        active_network.create_and_set_or_set_boa_env()

    if not active_network.save_to_db or not active_network.db_path:
        logger.error(
            f"Cannot get deployments without a database path on network {active_network.name}.\nPlease specify one or change networks."
        )
        return []

    deployments_list = []
    if not isinstance(limit, int) and not isinstance(limit, type(None)):
        raise ValueError(f"Limit must be an integer or None, not {type(limit)}.")
    if checked:
        deployments_list = active_network.get_deployments_checked(
            contract_name=contract_name, limit=limit, chain_id=active_network.chain_id
        )
    else:
        deployments_list = active_network.get_deployments_unchecked(
            contract_name=contract_name, limit=limit, chain_id=active_network.chain_id
        )

    int_format_level = int(format_level)
    if int_format_level > len(PrintVerbosity):
        int_format_level = PrintVerbosity.RAW.value

    print_deployments(deployments_list, PrintVerbosity(int_format_level))
    return deployments_list


def print_deployments(deployments_list: list[Deployment], format_level: PrintVerbosity):
    if len(deployments_list) > 0:
        print("-" * NUM_DASH)
    for deployment in deployments_list:
        if format_level == PrintVerbosity.CONTRACT_ADDRESS:
            print(f"Contract Address: {deployment.contract_address}")
            continue

        elif format_level == PrintVerbosity.CONTRACT_ADDRESS_AND_NAME:
            print(f"Contract Address: {deployment.contract_address}")
            print(f"Contract Name: {deployment.contract_name}")
            print(f"Deployer: {deployment.deployer}")
            print("-" * NUM_DASH)
            continue

        elif format_level == PrintVerbosity.ADDRESS_NAME_DEPLOYER:
            print(f"Contract Address: {deployment.contract_address}")
            print(f"Contract Name: {deployment.contract_name}")
            print(f"Deployer: {deployment.deployer}")
            print(f"RPC: {deployment.rpc}")
            print(f"Transaction Hash: {deployment.tx_hash}")
            print("Source Code:")
            for file_name, content in deployment.source_code["sources"].items():
                print(f"  File: {file_name}")
                print("  Content:")
                print(content["content"][:200] + "...")  # Print first 200 characters
            print("-" * NUM_DASH)
            continue

        elif format_level == PrintVerbosity.FULL_DETAILS:
            print(f"Contract Address: {deployment.contract_address}")
            print(f"Contract Name: {deployment.contract_name}")
            print(f"Deployer: {deployment.deployer}")
            print(f"RPC: {deployment.rpc}")
            print(f"Transaction Hash: {deployment.tx_hash}")
            print(f"Broadcast Timestamp: {deployment.broadcast_ts}")
            print("Transaction Dict:")
            for key, value in deployment.tx_dict.items():
                print(f"  {key}: {value}")
            print("Receipt Dict:")
            for key, value in deployment.receipt_dict.items():
                print(f"  {key}: {value}")
            print("Source Code:")
            for file_name, content in deployment.source_code["sources"].items():
                print(f"  File: {file_name}")
                print("  Content:")
                print(content["content"])
            print("ABI:")
            for item in deployment.abi:
                print(f"  {item}")
            print(f"Session ID: {deployment.session_id}")
            print(f"Deployment ID: {deployment.deployment_id}")
            print("-" * NUM_DASH)
            continue

        elif format_level == PrintVerbosity.RAW:
            print(deployment)
            print("-" * NUM_DASH)
            continue

    print(f"Total deployments: {len(deployments_list)}")
    return deployments_list



from moccasin.constants.vars import (
    DEFAULT_NETWORKS_BY_CHAIN_ID,
    DEFAULT_NETWORKS_BY_NAME,
)


def test_default_networks_by_name_match_chain_id():
    for chain_id, network in DEFAULT_NETWORKS_BY_CHAIN_ID.items():
        assert DEFAULT_NETWORKS_BY_NAME[network["name"]]["chain_id"] == chain_id
        assert DEFAULT_NETWORKS_BY_CHAIN_ID[chain_id]["name"] == network["name"]


def test_default_networks_by_chain_id_match_name():
    for name, network in DEFAULT_NETWORKS_BY_NAME.items():
        assert DEFAULT_NETWORKS_BY_CHAIN_ID[network["chain_id"]]["name"] == name
        assert DEFAULT_NETWORKS_BY_NAME[name]["chain_id"] == network["chain_id"]



import importlib.util
from argparse import Namespace
from pathlib import Path

from moccasin._sys_path_and_config_setup import (
    _patch_sys_path,
    _setup_network_and_account_from_config_and_cli,
    get_sys_paths_list,
)
from moccasin.config import get_config, initialize_global_config
from moccasin.logging import logger


def main(args: Namespace) -> int:
    initialize_global_config()
    run_script(
        args.script_name_or_path,
        network=args.network,
        account=args.account,
        private_key=args.private_key,
        password=args.password,
        password_file_path=args.password_file_path,
        fork=args.fork,
        url=args.url,
        prompt_live=args.prompt_live,
        db_path=args.db_path,
    )
    return 0


def run_script(
    script_name_or_path: Path | str,
    network: str = None,
    account: str = None,
    private_key: str = None,
    password: str = None,
    password_file_path: Path = None,
    fork: bool = False,
    url: str = None,
    prompt_live: bool = None,
    db_path: str = None,
):
    config = get_config()
    script_path: Path = get_script_path(script_name_or_path)

    # Set up the environment (add necessary paths to sys.path, etc.)
    with _patch_sys_path(get_sys_paths_list(config)):
        _setup_network_and_account_from_config_and_cli(
            network=network,
            url=url,
            fork=fork,
            account=account,
            private_key=private_key,
            password=password,
            password_file_path=password_file_path,
            prompt_live=prompt_live,
            db_path=db_path,
        )

        # We give the user's script the module name "deploy_script_moccasin"
        spec = importlib.util.spec_from_file_location(
            "deploy_script_moccasin", script_path
        )
        if spec is None:
            raise Exception(f"Cannot find script '{script_path}'")

        if spec.loader is None:
            raise Exception(f"not a module: '{script_path}'")

        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        # neat functionality:
        if hasattr(module, "moccasin_main") and callable(module.moccasin_main):
            result = module.moccasin_main()
            return result


def get_script_path(script_name_or_path: Path | str) -> Path:
    script_path = Path(script_name_or_path)
    root = get_config().get_root()
    if script_path.suffix != ".py":
        script_path = script_path.with_suffix(".py")

    if not script_path.is_absolute():
        if "script" not in script_path.parts:
            script_path = root / "script" / script_path
        else:
            script_path = root / script_path

    if not script_path.exists():
        logger.error(f"{script_path} not found")

    return script_path



import os
import subprocess
from pathlib import Path

EXPECTED_HELP_TEXT = "Vyper compiler"


def test_compile_help(mox_path):
    result = subprocess.run(
        [mox_path, "compile", "-h"], check=True, capture_output=True, text=True
    )
    assert (
        EXPECTED_HELP_TEXT in result.stdout
    ), "Help output does not contain expected text"
    assert result.returncode == 0


def test_build_help(mox_path):
    result = subprocess.run(
        [mox_path, "build", "-h"], check=True, capture_output=True, text=True
    )
    assert (
        EXPECTED_HELP_TEXT in result.stdout
    ), "Help output does not contain expected text"
    assert result.returncode == 0


def test_compile_alias_build_project(
    complex_temp_path, complex_cleanup_out_folder, mox_path
):
    current_dir = Path.cwd()
    try:
        os.chdir(current_dir.joinpath(complex_temp_path))
        result = subprocess.run(
            [mox_path, "build"], check=True, capture_output=True, text=True
        )
    finally:
        os.chdir(current_dir)
    assert "Running compile command" in result.stderr
    assert result.returncode == 0


def test_compile_one(complex_temp_path, complex_cleanup_out_folder, mox_path):
    current_dir = Path.cwd()
    try:
        os.chdir(current_dir.joinpath(complex_temp_path))
        result = subprocess.run(
            [mox_path, "build", "BuyMeACoffee.vy"],
            check=True,
            capture_output=True,
            text=True,
        )
    finally:
        os.chdir(current_dir)
    assert "Done compiling BuyMeACoffee" in result.stderr
    assert result.returncode == 0



from moccasin.boa_tools import VyperContract
from moccasin.config import get_config


# Test to make sure the same address is returned when calling
# manifest twice
def manifest_many_times():
    active_network = get_config().get_active_network()
    price_feed_one: VyperContract = active_network.get_or_deploy_contract("price_feed")
    print(price_feed_one.address)
    price_feed_two: VyperContract = active_network.manifest_contract("price_feed")
    print(price_feed_two.address)
    price_feed_three: VyperContract = active_network.get_or_deploy_contract(
        "price_feed"
    )
    print(price_feed_three.address)

    price_feed_different: VyperContract = active_network.get_or_deploy_contract(
        "price_feed", force_deploy=True
    )
    print(price_feed_different.address)


def moccasin_main():
    return manifest_many_times()



from pathlib import Path

import tomli_w

from moccasin.constants.vars import (
    DEFAULT_NETWORKS_BY_NAME,
    ERAVM,
    FORK_NETWORK_DEFAULTS,
    LOCAL_NETWORK_DEFAULTS,
    PYEVM,
)

AUTOGENERATED_DIR = Path(__file__).parent.joinpath("_autogenerated")

PYEVM_FILE = "pyevm_defaults.rst"
ERAVM_FILE = "eravm_defaults.rst"
FORKS_FILE = "fork_defaults.rst"
EXPLORER_FILE = "explorer_defaults.rst"


def clean_dict_for_toml(networks_dict):
    cleaned = {}
    for network, config in networks_dict.items():
        cleaned[network] = {k: ("" if v is None else v) for k, v in config.items()}
    return cleaned


if __name__ == "__main__":
    with open(AUTOGENERATED_DIR.joinpath(PYEVM_FILE), "wb") as f:
        defaults = LOCAL_NETWORK_DEFAULTS[PYEVM]
        tomli_w.dump(defaults, f)

    with open(AUTOGENERATED_DIR.joinpath(ERAVM_FILE), "wb") as f:
        defaults = LOCAL_NETWORK_DEFAULTS[ERAVM]
        tomli_w.dump(defaults, f)

    with open(AUTOGENERATED_DIR.joinpath(FORKS_FILE), "wb") as f:
        defaults = FORK_NETWORK_DEFAULTS
        tomli_w.dump(defaults, f)

    with open(AUTOGENERATED_DIR.joinpath(EXPLORER_FILE), "wb") as f:
        defaults = clean_dict_for_toml(DEFAULT_NETWORKS_BY_NAME)
        tomli_w.dump(defaults, f)
